import datetime as dt
import re
from pathlib import Path

import numpy as np
import pandas as pd

from energytool.tools import to_list


def eplus_date_parser(timestamp: str) -> dt.datetime:
    """Convert energyplus timestamp to datetime."""
    if not isinstance(timestamp, str):
        raise TypeError()

    timestamp = timestamp.lstrip()
    if " 24:" in timestamp:
        # EnergyPlus works with 1-24h and python with 0-23h
        timestamp = timestamp.replace(" 24:", " 23:")
        return dt.datetime.strptime(timestamp, "%m/%d %H:%M:%S")
    return dt.datetime.strptime(timestamp, "%m/%d %H:%M:%S") - dt.timedelta(hours=1)


def read_eplus_res(file_path: Path, ref_year: int = None):
    """
    Read EnergyPlus result data from output CSV file and adjust the date/time index.

    Parameters:
    - file_path (Path): The path to the EnergyPlus result file in CSV format.
    - ref_year (int, optional): The reference year for adjusting the date/time index.
      If not provided, the current year will be used as the reference year.

    Returns:
    - results (DataFrame): A pandas DataFrame containing the EnergyPlus result data
      with the adjusted date/time index.

    Raises:
    - ValueError: If the specified EnergyPlus result file is not found.
    """
    try:
        results = pd.read_csv(file_path, index_col=0)
    except FileNotFoundError:
        raise ValueError("EnergyPlus result file not found")
    results.index = [eplus_date_parser(ind) for ind in results.index]

    if ref_year is None:
        ref_year = dt.datetime.today().year

    timestep = results.index[1] - results.index[0]
    dt_range = pd.date_range(
        results.index[0].replace(year=int(ref_year)),
        periods=results.shape[0],
        freq=timestep,
    )
    dt_range.name = "Date/Time"
    results.index = dt_range

    return results


def zone_contains_regex(elmt_list):
    tempo = [elmt + ":.+|" for elmt in elmt_list]
    return "".join(tempo)[:-1]


def variable_contains_regex(elmt_list):
    if not elmt_list:
        return None
    tempo = [elmt + ".+|" for elmt in elmt_list]
    return "".join(tempo)[:-1]


def get_output_variable(
    eplus_res: pd.DataFrame,
    variables: str | list,
    key_values: str | list = "*",
    drop_suffix=True,
) -> pd.DataFrame:
    """
    This function allows you to extract specific output variables from an EnergyPlus
     result DataFrame based on the provided variable names and key values.

    :param eplus_res: A pandas DataFrame containing EnergyPlus simulation results.
        Index is a DateTimeIndex, columns are output variables
    :param variables: The names of the specific output variables to retrieve.
        This can be a single variable name (string) or a list of variable names
        (list of strings).
    :param key_values: (Optional) The key values that identify the simulation
        outputs. This can be a single key value (string) or a list of key values
        (list of strings). By default, "*" is used to retrieve variables for all
        key values.
    :param drop_suffix: (Optional) If True, remove the suffixes from the column
        names in the returned DataFrame. Default is True.
    :return: A DataFrame containing the selected output variables.
    Example:
    ```
    get_output_variable(
        eplus_res=toy_df,
        key_values="Zone1",
        variables="Equipment Total Heating Energy",
    )


    get_output_variable(
        eplus_res=toy_df,
        key_values=["Zone1", "ZONE2"],
        variables="Equipment Total Heating Energy",
    )

    get_output_variable(
        eplus_res=toy_df,
        key_values="*",
        variables="Equipment Total Heating Energy",
    )

    get_output_variable(
        eplus_res=toy_df,
        key_values="Zone1",
        variables=[
            "Equipment Total Heating Energy",
            "Ideal Loads Supply Air Total Heating Energy",
        ],
    )
    ```

    """
    if key_values == "*":
        key_mask = np.full((1, eplus_res.shape[1]), True).flatten()
    else:
        key_list = to_list(key_values)
        key_list_upper = [elmt.upper() for elmt in key_list]
        reg_key = zone_contains_regex(key_list_upper)
        key_mask = eplus_res.columns.str.contains(reg_key)

    variable_names_list = to_list(variables)
    reg_var = variable_contains_regex(variable_names_list)
    variable_mask = eplus_res.columns.str.contains(reg_var)

    mask = np.logical_and(key_mask, variable_mask)

    results = eplus_res.loc[:, mask]

    if drop_suffix:
        new_columns = [re.sub(f":{variables}.+", "", col) for col in results.columns]
        results.columns = new_columns

    return results



from typing import Any

from energytool.base.idf_utils import get_objects_name_list
from energytool.base.idfobject_utils import (
    get_windows_by_boundary_condition,
    get_constructions_layer_list,
)
from energytool.building import Building
from energytool.tools import is_items_in_list


def reverse_kwargs(construction_kwargs):
    construction_name = construction_kwargs["Name"]

    construction_layers = [
        value for key, value in construction_kwargs.items() if key != "Name" and value
    ]

    reversed_layers = construction_layers[::-1]

    reversed_kwargs = {"Name": construction_name, "Outside_Layer": reversed_layers[0]}
    reversed_kwargs.update(
        {f"Layer_{idx + 2}": layer for idx, layer in enumerate(reversed_layers[1:])}
    )

    return reversed_kwargs


def set_opaque_surface_construction(
    model: Building,
    description: dict[str, list[dict[str, Any]]],
    name_filter: str = None,
    surface_type: str = "Wall",
    outside_boundary_condition: str = None,
):
    """
    This function modifies the construction of opaque building surfaces in an EnergyPlus IDF file.
    It is intended for use as a modifier for the corrai variant framework.

    :param model: energytool Building object.
    :param description: A dictionary describing the construction
    materials and properties. The argument must be of the form:
        {
            "construction_name": [
                {
                    "Name": "material_1_name",
                    "Thickness": 0.01,
                    ...
                },
                {
                    "Name": "material_2_name",
                    "Thickness": 0.5,
                    ...
                }
            ]
        }
    :param name_filter: An optional filter for surface names.
    :param surface_type: The type of surface to modify (default is 'Wall').
    :param outside_boundary_condition: The outside boundary condition (default is None).

    This function first identifies the surfaces to modify based on the provided parameters.
    It then modifies the construction of these surfaces according
    to the provided construction description.

    If a new construction is created during this process, its properties are stored in kwargs.
    These kwargs are then reversed to ensure consistency for any surfaces
    that require the inversion of their construction.
    """
    if name_filter is None:
        name_filter = ""

    new_construction_name = list(description.keys())[0]
    new_composition = description[new_construction_name]
    surface_list = model.idf.idfobjects["BuildingSurface:Detailed"]

    if all(
        surface_type not in obj.getfieldidd("Surface_Type")["key"]
        for obj in surface_list
    ):
        raise ValueError(
            f"surface_type must be one of "
            f"{surface_list[0].getfieldidd('Surface_Type')['key']}, "
            f"got {surface_type}"
        )

    if outside_boundary_condition is not None:
        if any(
            outside_boundary_condition
            not in obj.getfieldidd("Outside_Boundary_Condition")["key"]
            for obj in surface_list
            if name_filter in obj.Name
        ):
            raise ValueError(
                f"outside_boundary_condition must be one of "
                f"{surface_list[0].getfieldidd('Outside_Boundary_Condition')['key']}, "
                f"got {outside_boundary_condition}"
            )

    for material in new_composition:
        if "Roughness" not in material.keys():
            material["Roughness"] = "Rough"

        if material["Name"] not in get_objects_name_list(model.idf, "Material"):
            model.idf.newidfobject("Material", **material)

    construction_kwargs = {
        "Name": new_construction_name,
        "Outside_Layer": new_composition[0]["Name"],
    }

    if len(new_composition) > 1:
        for idx, mat in enumerate(new_composition[1:]):
            construction_kwargs[f"Layer_{idx + 2}"] = mat["Name"]

    if new_construction_name not in get_objects_name_list(model.idf, "Construction"):
        model.idf.newidfobject("Construction", **construction_kwargs)

    surf_to_modify = [
        obj
        for obj in surface_list
        if obj.Surface_Type == surface_type
        and (
            obj.Outside_Boundary_Condition == outside_boundary_condition
            if outside_boundary_condition is not None
            else True
        )
        and name_filter in obj.Name
    ]

    for surf in surf_to_modify:
        surf.Construction_Name = new_construction_name

    construction_to_reverse = [
        obj.Construction_Name
        for obj in surface_list
        if name_filter in obj.Outside_Boundary_Condition_Object
        and (
            getattr(
                obj.Outside_Boundary_Condition_Object,
                "Outside_Boundary_Condition",
                None,
            )
            == outside_boundary_condition
            if outside_boundary_condition is not None
            else True
        )
        and obj.Surface_Type == surface_type
    ]

    reversed_kwargs = reverse_kwargs(construction_kwargs)
    for construction_name in construction_to_reverse:
        construction_object = model.idf.getobject("Construction", construction_name)
        num_layers = min(len(reversed_kwargs), len(construction_object.fieldnames) - 1)
        for idx, (_, value) in enumerate(reversed_kwargs.items()):
            if idx < num_layers:
                construction_object[construction_object.fieldnames[idx + 1]] = value
            else:
                break
        construction_object["Name"] = construction_name

        for field in construction_object.fieldnames[num_layers + 1 :]:
            if field in construction_object:
                construction_object.pop(field)


def set_external_windows(
    model: Building,
    description: dict[str, dict[str, Any]],
    name_filter: str = None,
    surface_name_filter: str = None,
    boundary_conditions: str = None,
):
    """
    Replace windows in an EnergyPlus building model with new window descriptions.

    This function iterates through the windows in the model, filters them based on their
    name and boundary conditions, and replaces them with new window descriptions.
    It also handles associated constructions and materials.

    Parameters:
    :param model: An EnergyPlus building model.
    :param description: A dictionary containing the new window description(s).
        The expected dictionary must be of the following form:
        {
            "Variant_1": {
                "Name": "Var_1",
                "UFactor": 1,
                "Solar_Heat_Gain_Coefficient": 0.1,
                "Visible_Transmittance": 0.1,
            },
        }
    :param name_filter: An optional filter to match window names.
    :param surface_name_filter: An optional filter to match window surface names.
    :param boundary_conditions: The boundary condition for the windows
        (default is "Outdoors").

    """
    idf = model.idf

    # Get windows materials list and shaded windows constructions
    if boundary_conditions:
        windows = get_windows_by_boundary_condition(
            idf, boundary_condition=boundary_conditions
        )
    else:
        windows = idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]

    if name_filter is not None or surface_name_filter is not None:
        windows = [
            win
            for win in windows
            if (
                name_filter is None and surface_name_filter in win.Building_Surface_Name
            )
            or (surface_name_filter is None and name_filter in win.Name)
        ]
    windows_names = [win.Name for win in windows]
    win_cons_names = {win.Construction_Name for win in windows}
    windows_constructions = [
        idf.getobject("Construction", name) for name in win_cons_names
    ]

    win_mat_list = get_constructions_layer_list(windows_constructions)
    windows_materials = [
        idf.getobject("WindowMaterial:SimpleGlazingSystem", name)
        for name in set(win_mat_list)
        if idf.getobject("WindowMaterial:SimpleGlazingSystem", name) is not None
    ]

    shading_controls = [
        obj
        for obj in idf.idfobjects["WindowShadingControl"]
        if any(is_items_in_list(obj.fieldvalues, windows_names))
    ]

    obj_list = [
        obj.get_referenced_object("Construction_with_Shading_Name")
        for obj in shading_controls
    ]
    set_name = {obj.Name for obj in obj_list}

    shaded_window_constructions = [
        idf.getobject("Construction", name) for name in set_name
    ]

    # Replace windows
    new_window_name = list(description.keys())[0]
    new_window = description[new_window_name]

    name_to_replace = [obj.Name for obj in windows_materials]
    constructions_list = windows_constructions + shaded_window_constructions
    for construction in constructions_list:
        for field in construction.fieldnames:
            if construction[field] in name_to_replace:
                construction[field] = new_window["Name"]

    # Add the new window material to the IDF
    if new_window["Name"] not in [
        win.Name for win in idf.idfobjects["WindowMaterial:SimpleGlazingSystem"]
    ]:
        idf.newidfobject(key="WindowMaterial:SimpleGlazingSystem", **new_window)

    used_mat_list = [
        val for cons in idf.idfobjects["CONSTRUCTION"] for val in cons.fieldvalues[2:]
    ]

    idf.idfobjects["WindowMaterial:SimpleGlazingSystem"] = [
        win
        for win in idf.idfobjects["WindowMaterial:SimpleGlazingSystem"]
        if win.Name in used_mat_list
    ]


def set_afn_surface_opening_factor(
    model: Building,
    description: dict[str, dict[str, Any]],
    name_filter: str = None,
    surface_name_filter: str = None,
):
    """
    Modify AirFlowNetwork:Multizone:Surface WindowDoor_Opening_Factor_or_Crack_Factor
    based on their name.

    :param model: An EnergyPlus building model.
    :param description: A dictionary containing the new value.
        the expected dictionary must be of the following form:
        {
            "Variant_1": {
                "WindowDoor_Opening_Factor_or_Crack_Factor": 0.3,
            },
        }
    :param name_filter: An optional filter to match window names.
    :param surface_name_filter: An optional filter to match window surface names.
    """
    idf = model.idf

    openings = idf.idfobjects["AirflowNetwork:MultiZone:Surface"]

    if name_filter is not None or surface_name_filter is not None:
        openings = [
            op
            for op in openings
            if (surface_name_filter is None and name_filter in op.Surface_Name)
            or (name_filter is None and surface_name_filter in op.Surface_Name)
        ]

    new_opening_ratio_name = list(description.keys())[0]
    new_opening_ratio = description[new_opening_ratio_name][
        "WindowDoor_Opening_Factor_or_Crack_Factor"
    ]

    for opening in openings:
        opening["WindowDoor_Opening_Factor_or_Crack_Factor"] = new_opening_ratio


def set_blinds_solar_transmittance(
    model: Building,
    description: dict[str, dict[str, Any]],
    name_filter: str = None,
    surface_name_filter: str = None,
):
    """
    Modify WindowMaterial:Shade Solar_Transmittance (or/and Reflectance) based
    on the given description.

    :param model: An EnergyPlus building model.
    :param description: A dictionary containing the new values for shades.
        The expected dictionary must be of the following form:
        {
            "Variant_1": [
                {
                    "Solar_Transmittance": 0.66,
                    "Solar_Reflectance": 0.20
                }
            ]
        }
    :param name_filter: An optional filter to match window names.
    :param surface_name_filter: An optional filter to match window surface names.
    """
    idf = model.idf

    shades = idf.idfobjects["WindowMaterial:Shade"]
    all_constructions = idf.idfobjects["Construction"]
    scenarios = idf.idfobjects["WindowShadingControl"]

    new_shaded_window_name = list(description.keys())[0]

    selected_shades = []

    if name_filter is None:
        name_filter = ""
    if surface_name_filter is None:
        surface_name_filter = ""

    filtered_windows = [
        window
        for window in idf.idfobjects["FenestrationSurface:Detailed"]
        if (surface_name_filter == "" and name_filter in window.Name)
        or (name_filter == "" and surface_name_filter in window.Building_Surface_Name)
        or (surface_name_filter == "" and name_filter == "")
        or (
            surface_name_filter in window.Building_Surface_Name
            and name_filter in window.Name
        )
    ]
    construction_names_dict = {
        window.Name: window.Construction_Name for window in filtered_windows
    }

    # Check if construction_name of filtered window includes a shade or a shaded version
    for window_name, target_name in construction_names_dict.items():
        for construction in all_constructions:
            if (
                construction.Name == target_name
                or construction.Name == target_name + "_Shaded"
            ):
                construction_values = [
                    construction[field] for field in construction.fieldnames[2:]
                ]
                for shade in shades:
                    if any(
                        construction_value == shade.Name
                        for construction_value in construction_values
                        if construction_value
                    ):
                        selected_shades.append(shade)

        # Also, check "WINDOWSHADINGCONTROL" construction associated to filtered windows
        for scen in scenarios:
            for n in range(1, 10):
                if scen[f"Fenestration_Surface_{n}_Name"] == window_name:
                    construction_name = scen.Construction_with_Shading_Name
                    for construction in all_constructions:
                        if construction.Name == construction_name:
                            construction_values = [
                                construction[field]
                                for field in construction.fieldnames[2:]
                            ]
                            for shade in shades:
                                if any(
                                    construction_value == shade.Name
                                    for construction_value in construction_values
                                    if construction_value
                                ):
                                    selected_shades.append(shade)

    new_transmittance = description[new_shaded_window_name][0].get(
        "Solar_Transmittance"
    )
    new_reflectance = description[new_shaded_window_name][0].get("Solar_Reflectance")

    for shade in selected_shades:
        if new_transmittance is not None:
            shade["Solar_Transmittance"] = new_transmittance
        if new_reflectance is not None:
            shade["Solar_Reflectance"] = new_reflectance


def set_schedule_constant(
    model: Building,
    description: dict[str, dict[str, Any]],
):
    idf = model.idf

    schedule_constant = idf.idfobjects["SCHEDULE:CONSTANT"]

    for schedule_name, schedule_fields in description.items():
        schedule_exists = False

        for sched in schedule_constant:
            if sched["Name"] == schedule_fields["Name"]:
                sched["Hourly_Value"] = schedule_fields["Hourly_Value"]
                schedule_exists = True
                break  # Exit loop once found and modified

        if not schedule_exists:
            new_schedule = {
                "Name": schedule_fields["Name"],
                "Schedule_Type_Limits_Name": schedule_fields[
                    "Schedule_Type_Limits_Name"
                ],
                "Hourly_Value": schedule_fields["Hourly_Value"],
            }
            model.idf.newidfobject("SCHEDULE:CONSTANT", **new_schedule)


def set_blinds_schedule(
    model: Building,
    description: dict[str, dict[str, Any]],
    name_filter: str = None,
    surface_name_filter: str = None,
):
    """
    Create/update Schedule based on the given description.

    :param model: An EnergyPlus building model.
    :param description: A dictionary containing the new values for schedule.
        The expected dictionary must be of the following form:
        {
            "Variant_1": [
                {
                    "Scenario": {
                        "Name": 'Shading_control',
                        "Schedule_Type_Limits_Name": 'Fractional1',
                        "Field1": "Through: 01 April",
                        # ... other fields ...
                    },
                    "Limits": {
                        "Name": 'Fractional1',
                        "Lower_Limit_Value": 0,
                        "Upper_Limit_Value": 1,
                        "Numeric_Type": "Continuous"
                    }
                }
            ]
        }
    :param name_filter: An optional filter to match window names.
    :param surface_name_filter: An optional filter to match window surface names.
    """
    idf = model.idf

    scenarios = idf.idfobjects["WindowShadingControl"]
    new_shaded_window_name = list(description.keys())[0]
    new_schedule = description[new_shaded_window_name][0]["Scenario"]

    schedule = idf.idfobjects["Schedule:Year"]
    compact = idf.idfobjects["Schedule:Compact"]
    existing_shading_control = [entry["Name"] for entry in schedule] + [
        entry["Name"] for entry in compact
    ]

    name_in_existing = any(
        new_schedule["Name"] in entry for entry in existing_shading_control
    )

    if not name_in_existing and "Field_1" not in new_schedule:
        raise ValueError(
            "Scenario's name not found in IDF. "
            "Use existing name or define Schedule "
            "fields in description"
        )

    if name_filter or surface_name_filter:
        filtered_windows = [
            window
            for window in idf.idfobjects["FenestrationSurface:Detailed"]
            if (surface_name_filter is None and name_filter in window.Name)
            or (
                name_filter is None
                and surface_name_filter in window.Building_Surface_Name
            )
        ]
        construction_names_dict = {
            window.Name: window.Construction_Name for window in filtered_windows
        }
    else:
        construction_names_dict = {
            window.Name: window.Construction_Name
            for window in idf.idfobjects["FenestrationSurface:Detailed"]
        }

    for wind_name, _ in construction_names_dict.items():
        for scen in scenarios:
            for n in range(1, 10):
                # check if construction_Name matches construction names of windows +
                # check if window is assigned to Fenestration_Surface_1 to _10
                # of "WINDOWSHADINGCONTROL"
                if scen[f"Fenestration_Surface_{n}_Name"] == wind_name:
                    scen["Schedule_Name"] = new_schedule["Name"]

    required_fields = ["Name", "Schedule_Type_Limits_Name"]

    if any(field not in required_fields for field in new_schedule):
        # More than Name or Schedule_Type_Limits_Name is given in Description
        schedule_kwargs = {
            "Name": new_schedule["Name"],
            "Schedule_Type_Limits_Name": (
                new_schedule["Schedule_Type_Limits_Name"]
                if "Schedule_Type_Limits_Name" in new_schedule
                else "Fractional"
            ),
        }

        for idx, info in enumerate(new_schedule.values()):
            if "Schedule_Type_Limits_Name" in new_schedule.keys():
                if idx >= 2:
                    schedule_kwargs[f"Field_{idx - 1}"] = info
            else:
                if idx >= 1:
                    schedule_kwargs[f"Field_{idx}"] = info

        existing_st_limits = [
            entry["Name"] for entry in idf.idfobjects["ScheduleTypeLimits"]
        ]
        if (
            "Limits" not in description[new_shaded_window_name][0]
            and schedule_kwargs["Schedule_Type_Limits_Name"] not in existing_st_limits
        ):
            raise ValueError(
                "ScheduleTypeLimit is not specified in IDF. Define "
                "ScheduleTypeLimit fields in Description"
            )

        model.idf.newidfobject(
            "Schedule:Compact", **schedule_kwargs
        )  # new or replaced ?

        st_limit = schedule_kwargs["Schedule_Type_Limits_Name"]
        existing_st_limits = [
            entry["Name"] for entry in idf.idfobjects["ScheduleTypeLimits"]
        ]

        if st_limit not in existing_st_limits and (
            limits := description.get(new_shaded_window_name, [{}])[0].get("Limits")
        ):
            limits_kwargs = {
                "Name": new_schedule["Schedule_Type_Limits_Name"],
                "Lower_Limit_Value": limits["Lower_Limit_Value"],
                "Upper_Limit_Value": limits["Upper_Limit_Value"],
                "Numeric_Type": limits["Numeric_Type"],
            }
            model.idf.newidfobject("ScheduleTypeLimits", **limits_kwargs)


def update_idf_objects(
    model: Building,
    description: dict[str, dict[str, Any]],
    idfobject_type: str,
    name_filter: str = None,
):
    """
    Updates or creates objects in an IDF based on the provided description.

    This function updates the fields of existing objects in an IDF or creates new objects
    if no matching objects are found. A partial name filter can be used to update only
    the objects whose names contain the specified filter.

    Parameters:
    model (Building): The building model containing the IDF.
    description (dict[str, dict[str, Any]]): A dictionary describing the objects to be updated or created.
        Example:
            description = {
                "Schedule1": {
                    "Name": "Schedule_test1",
                    "Schedule_Type_Limits_Name": "Fractional",
                    "Hourly_Value": 0.77,
                },
            }
    idfobject_type (str): The type of IDF object to be updated or created.
    name_filter (str, optional): A partial name filter to match objects for updating. Defaults to None.

    """
    idf = model.idf
    idf_objects = idf.idfobjects[idfobject_type]

    for obj_name, obj_fields in description.items():
        obj_name_filter = obj_fields.get("Name")
        obj_exists = False

        for obj in idf_objects:
            if name_filter is not None and name_filter in obj["Name"]:
                for field, value in obj_fields.items():
                    if field != "Name":
                        obj[field] = value
                obj_exists = True
            elif name_filter is None and obj["Name"] == obj_name_filter:
                for field, value in obj_fields.items():
                    if field != "Name":
                        obj[field] = value
                obj_exists = True

        if not obj_exists and name_filter is None:
            new_obj_kwargs = {field: value for field, value in obj_fields.items()}
            model.idf.newidfobject(idfobject_type, **new_obj_kwargs)


def set_blinds_st_and_schedule(
    model: Building,
    description: dict[str, dict[str, Any]],
    name_filter: str = None,
    surface_name_filter: str = None,
):
    """
    Modify WindowMaterial:Shade Solar_Transmittance and create/update
    Schedule based on the given description.

    :param model: An EnergyPlus building model.
    :param description: A dictionary containing the new values for shades and schedule.
        The expected dictionary must be of the following form:
        {
            "Variant_1": [
                {
                    "Solar_Transmittance": 0.66,
                    "Scenario": {
                        "Name": 'Shading_control',
                        "Schedule_Type_Limits_Name": 'Fractional1',
                        "Field1": "Through: 01 April",
                        "Field2": "For: AllDays",
                        "Field4": "Until: 24:00",
                        "Field3": 0.0,
                        "Field5": "Through: 30 September",
                        "Field6": "For: Weekdays",
                        "Field7": "Until: 24:00",
                        "Field8": 1.0,
                        "Field9": "For: Weekends",
                        "Field10": "Until: 24:00",
                        "Field11": 0.0,
                        "Field12": "For: AllOtherDays",
                        "Field13": "Until: 24:00",
                        "Field14": 0.0,
                    },
                    "Limits": {
                        "Name": 'Fractional1',
                        "Lower_Limit_Value": 0,
                        "Upper_Limit_Value": 1,
                        "Numeric_Type": "Continuous"
                    }
                }
            ]
        }
    :param name_filter: An optional filter to match window names.
    :param surface_name_filter: An optional filter to match window surface names.
    """
    set_blinds_solar_transmittance(model, description, name_filter, surface_name_filter)
    set_blinds_schedule(model, description, name_filter, surface_name_filter)


def set_system(model, description, **kwargs):
    system = list(description.values())[0]
    system_name = kwargs.get("system_name", system.name)

    # Remove existing system with the same name if needed
    model.del_system(system_name)

    # Add new system
    model.add_system(system)



import pandas as pd
import datetime as dt


def to_list(f_input):
    """
    Convert a string into a list
    return f_input if f_input is a list
    else raise ValueError

    :param f_input:
    :return: list
    """
    if isinstance(f_input, (str, int, float)):
        return [f_input]
    elif isinstance(f_input, list):
        return f_input
    else:
        raise ValueError(
            f"{f_input} must be an instance of str, int, float or list."
            f"Got {type(f_input)} instead"
        )


def select_in_list(target_list: list, target: str | list):
    """
    Select elements from a list based on a target string or a list of target strings.

    :param target_list: The source list from which elements will be selected.
    :param target: A string or a list of strings to match against elements in
    the target_list. If "*", all elements in the target_list are returned.

    :return: A list containing the selected elements from the target_list.
    """
    select_by_list = to_list(target)

    if target == "*":
        return target_list

    output_list = []
    for elmt in select_by_list:
        for items in target_list:
            if elmt in items:
                output_list.append(items)

    return output_list


def hourly_lst_from_dict(hourly_dict):
    if list(hourly_dict.keys())[-1] != 24:
        raise ValueError("Last dict key must be 24")

    val_list = []
    for hour, val in hourly_dict.items():
        val_list += [val for _ in range(len(val_list), hour)]

    return val_list


def is_items_in_list(items: str | list, target_list: list):
    """
    This function checks whether one or more items (strings or lists) are present
    within the target list.

    :param target_list: The list to search for items.
    :param items: The item(s) to check for presence in the target list. This can be a
    single string or a list of strings.
    :return: A list of Boolean values indicating whether each item is present in the
    target list.
    """
    items = to_list(items)
    return [True if elmt in target_list else False for elmt in items]


class Scheduler:
    def __init__(self, name, year=None):
        self.name = name
        if year is None:
            year = dt.datetime.today().year
        self.year = year
        self.series = pd.Series(
            index=pd.date_range(f"{year}-01-01 00:00:00", freq="h", periods=8760),
            name=name,
            dtype="float64",
        )

    def add_day_in_period(self, start, end, days, hourly_dict):
        start = dt.datetime.strptime(start, "%Y-%m-%d")
        end = dt.datetime.strptime(end, "%Y-%m-%d")
        end = end.replace(hour=23)

        if start.year != self.year or end.year != end.year:
            raise ValueError("start date or end date is out of bound ")

        day_list = to_list(days)
        period = self.series.loc[start:end]

        selected_timestamp = [idx for idx in period.index if idx.day_name() in day_list]

        self.series.loc[selected_timestamp] = hourly_lst_from_dict(hourly_dict) * int(
            len(selected_timestamp) / 24
        )



import enum
import json

from abc import ABC, abstractmethod
from pathlib import Path

import eppy.modeleditor
import numpy as np
import pandas as pd
from corrai.base.utils import as_1_column_dataframe
from eppy.modeleditor import IDF

import energytool.base.parse_results
from energytool.base.idf_utils import (
    get_objects_name_list,
    set_named_objects_field_values,
    del_named_objects,
)
from energytool.base.idfobject_utils import (
    get_zones_idealloadsairsystem,
    add_output_variable,
    get_number_of_people,
    add_hourly_schedules_from_df,
    add_natural_ventilation,
    add_obj_from_obj_dict,
)
from energytool.base.parse_results import get_output_variable
from energytool.base.units import Units
from energytool.tools import select_in_list, to_list

resource_path = Path(__file__).parent / "resources/resources.json"
with resource_path.open("r", encoding="utf-8") as f:
    RESOURCE_JSON = json.load(f)


class SystemCategories(enum.Enum):
    HEATING = "HEATING"
    COOLING = "COOLING"
    VENTILATION = "VENTILATION"
    LIGHTING = "LIGHTING"
    AUXILIARY = "AUXILIARY"
    DHW = "DHW"
    PV = "PV"
    SENSOR = "SENSOR"
    OTHER = "OTHER"


class System(ABC):
    def __init__(self, name: str, category: SystemCategories = SystemCategories.OTHER):
        self.name = name
        self.category = category

    def __repr__(self):
        return f"{self.name}"

    @abstractmethod
    def pre_process(self, idf: IDF):
        """Operations happening before the simulation"""
        pass

    @abstractmethod
    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        """Operations happening after the simulation"""
        pass


class Overshoot28(System):
    """
    Represents the thermal discomfort system for zones exceeding a temperature threshold.

    :param name: Name of the system.
    :param zones: Zones or spaces to monitor. Can be "*" for all zones or a list of zones.
    :param temp_threshold: Temperature threshold for thermal discomfort.
    :param occupancy_in_output: Set to True for Zone People Occupant Count in output df.
    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        temp_threshold: float = 28.0,
        occupancy_in_output=False,
    ):
        super().__init__(name=name, category=SystemCategories.SENSOR)
        self.zones = zones
        self.temp_threshold = temp_threshold
        self.occupancy_in_output = occupancy_in_output

    def pre_process(self, idf: IDF):
        """
        Adds the necessary output variables to monitor operative temperature and occupancy.
        """
        # Add output variable for Zone Operative Temperature
        add_output_variable(
            idf=idf,
            key_values=self.zones,
            variables="Zone Operative Temperature",
        )

        # Add output variable for Zone People Occupant Count
        add_output_variable(
            idf=idf,
            key_values=self.zones,
            variables="Zone People Occupant Count",
        )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        """
        Calculates thermal discomfort based on operative temperature and occupancy.

        :param idf: The EnergyPlus input data (not used in this method).
        :param eplus_results: DataFrame containing EnergyPlus simulation results.
        :return: DataFrame with additional columns for thermal discomfort.
        """
        operative_temperature = get_output_variable(
            eplus_res=eplus_results,
            key_values=self.zones,
            variables="Zone Operative Temperature",
        )

        occupancy = get_output_variable(
            eplus_res=eplus_results,
            key_values=self.zones,
            variables="Zone People Occupant Count",
        )

        results = pd.DataFrame(index=operative_temperature.index)

        for col in operative_temperature.columns:
            results[f"discomfort_{col}"] = (
                (operative_temperature[col] >= self.temp_threshold)
                & (occupancy[col] > 0)
            ).astype(int)
            if self.occupancy_in_output:
                results[f"occupancy_{col}"] = occupancy[col]

        return results


class LightAutonomy(System):
    """
    Represents a system to calculate lighting autonomy per zone.

    :param name: Name of the system.
    :param zones: Zones or spaces to monitor. Can be "*" for all zones or a list of zones.
    :param lux_threshold: Minimum lux level required for lighting autonomy.
    :param light_schedule_name: Name of the light schedule to retrieve from "Schedule Value".
    :param occupancy_in_output: Set to True for Zone People Occupant Count in output df.
    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        lux_threshold: float = 200,
        light_schedule_name: str = None,
        occupancy_in_output=False,
    ):
        super().__init__(name=name, category=SystemCategories.SENSOR)
        self.zones = zones
        self.lux_threshold = lux_threshold
        self.light_schedule_name = light_schedule_name
        self.occupancy_in_output = occupancy_in_output

    def pre_process(self, idf: IDF):
        """
        Adds the necessary output variables for illuminance, occupancy, and light schedule.

        :param idf: The EnergyPlus input data.
        """
        # Add output variable for Daylighting Reference Point Illuminance

        # Add output variable for the specified light schedule
        add_output_variable(
            idf=idf,
            key_values=self.light_schedule_name,
            variables="Schedule Value",
        )

        add_output_variable(
            idf=idf,
            key_values=self.zones,
            variables="Daylighting Reference Point 1 Illuminance",
        )

        # Add output variable for Zone People Occupant Count
        add_output_variable(
            idf=idf,
            key_values=self.zones,
            variables="Zone People Occupant Count",
        )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        """
        Calculates lighting autonomy based on illuminance, occupancy, and schedule.

        :param idf: The EnergyPlus input data (not used in this method).
        :param eplus_results: DataFrame containing EnergyPlus simulation results.
        :return: DataFrame with additional columns for lighting autonomy per zone.
        """
        schedule = get_output_variable(
            eplus_res=eplus_results,
            key_values=self.light_schedule_name,
            variables="Schedule Value",
        )

        # Get the occupancy data
        occupancy = get_output_variable(
            eplus_res=eplus_results,
            key_values=self.zones,
            variables="Zone People Occupant Count",
        )

        # Get the illuminance data
        illuminance = get_output_variable(
            eplus_res=eplus_results,
            key_values=self.zones,
            variables="Daylighting Reference Point 1 Illuminance",
        )

        # Calculate lighting autonomy for each zone
        is_occupied = occupancy > 0
        is_schedule_active = schedule > 0

        results = pd.DataFrame(index=illuminance.index)
        for col in illuminance.columns:
            results[f"autonomy_{col}"] = (
                (illuminance[col] >= self.lux_threshold)
                & is_occupied[col]
                & is_schedule_active.squeeze()
            ).astype(int)
            if self.occupancy_in_output:
                results[f"occupancy_{col}"] = occupancy[col]

        return results


class Sensor(System):
    """
    Add output:variables to the idf, get the results in post process.
    :param name(str): Name of the sensor
    :param variables: The names of the variables to output.
    :param key_values: The key values for which to add output variables.
        This can be a single key value (string) or a list of key values
        (list of strings). Default is '*' meaning all the available variables.
    """

    def __init__(self, name: str, variables: str, key_values: str | list[str] = "*"):
        super().__init__(name=name, category=SystemCategories.SENSOR)
        self.variables = variables
        self.key_values = key_values

    def pre_process(self, idf: IDF):
        add_output_variable(
            idf=idf,
            key_values=self.key_values,
            variables=self.variables,
        )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        results = get_output_variable(
            eplus_res=eplus_results,
            key_values=self.key_values,
            variables=self.variables,
        )
        results.columns = results.columns + f"_{self.variables}"
        return results


class SimplifiedChiller(System):
    """
    Represent a simplified chilling system with a coefficient of performance COP.
    The class is based on IdealLoadsAirSytem. For each provided zones, it will get the
    "Zone Ideal Loads Supply Air Total Cooling Energy" result and divide it by the cop.

    :parameter name(str): name of the system
    :parameter zone(str): idf zones controlled by the system. It must match zones in
        the idf file
    heated by the IdealLoadsAirSystem
    :parameter cop(float): Coefficient of Performance of the System. Can range from 0
        to +infinity

    attribute : category(SystemCategories): SystemCategories.COOLING

    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        cop=2.5,
    ):
        super().__init__(name=name, category=SystemCategories.COOLING)
        self.cop = cop
        self.zones = zones
        self.ilas_list = []

    def pre_process(self, idf: IDF):
        self.ilas_list = get_zones_idealloadsairsystem(idf, self.zones)

        add_output_variable(
            idf=idf,
            key_values=[ilas.Name for ilas in self.ilas_list],
            variables="Zone Ideal Loads Supply Air Total Cooling Energy",
        )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        # Warning, works only if ilas name contains zone name
        ideal_cooling = get_output_variable(
            eplus_res=eplus_results,
            key_values=[ilas.Name for ilas in self.ilas_list],
            variables="Zone Ideal Loads Supply Air Total Cooling Energy",
        )

        system_out = (ideal_cooling / self.cop).sum(axis=1)
        system_out.name = f"{self.name}_{Units.ENERGY.value}"
        return system_out.to_frame()


class HeaterSimple(System):
    """
    Represent a simple heating system with a coefficient of performance COP.
    The class is based on IdealLoadsAirSytem. For each provided zones, it will get the
    "Zone Ideal Loads Supply Air Total Heating Energy" result and divide it by the cop.

    :parameter name(str): name of the system
    :parameter zone(str): idf zones controlled by the system. It must match zones in
        the idf file
    heated by the IdealLoadsAirSystem
    :parameter cop(float): Coefficient of Performance of the System. Can range from 0
        to +infinity

    attribute : category(SystemCategories): SystemCategories.HEATING

    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        cop=1,
    ):
        super().__init__(name=name, category=SystemCategories.HEATING)
        self.cop = cop
        self.zones = zones
        self.ilas_list = []

    def pre_process(self, idf: IDF):
        self.ilas_list = get_zones_idealloadsairsystem(idf, self.zones)

        add_output_variable(
            idf=idf,
            key_values=[ilas.Name for ilas in self.ilas_list],
            variables="Zone Ideal Loads Supply Air Total Heating Energy",
        )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        # Warning, works only if ilas name contains zone name
        ideal_heating = get_output_variable(
            eplus_res=eplus_results,
            key_values=[ilas.Name for ilas in self.ilas_list],
            variables="Zone Ideal Loads Supply Air Total Heating Energy",
        )

        system_out = (ideal_heating / self.cop).sum(axis=1)
        system_out.name = f"{self.name}_{Units.ENERGY.value}"
        return system_out.to_frame()


class HeatingAuxiliary(System):
    """
    A simple way to model heating system auxiliary consumption as a ratio of the total
    heating needs.
    The class is based on IdealLoadsAirSytem. For each provided zones, it will get the
    "Zone Ideal Loads Supply Air Total Heating Energy" result and multiply it by a
     ratio.

    :parameter name(str): name of the system
    :parameter zone(str): idf zones controlled by the system. It must match zones in
        the idf file
    heated by the IdealLoadsAirSystem
    :parameter ratio(float): The ratio of auxiliary consumption. Can range from 0 to
        +infinity

    attribute : category(SystemCategories): SystemCategories.AUXILIARY

    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        ratio=0.05,
    ):
        super().__init__(name=name, category=SystemCategories.AUXILIARY)
        self.ratio = ratio
        self.zones = zones
        self.ilas_list = []

    def pre_process(self, idf: IDF):
        self.ilas_list = get_zones_idealloadsairsystem(idf, self.zones)

        add_output_variable(
            idf=idf,
            key_values=[ilas.Name for ilas in self.ilas_list],
            variables="Zone Ideal Loads Supply Air Total Heating Energy",
        )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        # Warning, works only if ilas name contains zone name
        ideal_heating = get_output_variable(
            eplus_res=eplus_results,
            key_values=[ilas.Name for ilas in self.ilas_list],
            variables="Zone Ideal Loads Supply Air Total Heating Energy",
        )

        system_out = (ideal_heating * self.ratio).sum(axis=1)
        system_out.name = f"{self.name}_{Units.ENERGY.value}"
        return system_out.to_frame()


class AirHandlingUnit(System):
    """
    A simple model for single flow and crossflow air handling units.
    This class is based on DesignSpecification:OutdoorAir objects and provides
    a convenient way to estimate fan energy consumption, set airflow using
    air changes per hour (ACH), and define heat recovery efficiency.

    Parameters:
        name (str): The name of the air handling unit.
        zones (str | List[str]): The name(s) of the zones served by the unit.
        fan_energy_coefficient (float): The fan energy coefficient in Wh/m3,
            used for fan energy consumption estimation.
        ach (float): The air change rate per hour in volume per hour (Vol/h).

    Notes:
    - If you use the "ach" argument, ensure that the DesignSpecification:OutdoorAir
      objects' names corresponding to the specified "zones" contain the zone names
      in their "Name" field.

    - Heat recovery efficiency settings will impact the latent and sensible
      efficiency of the heat exchanger between extracted and blown air.
    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        fan_energy_coefficient: float = 0.23,
        heat_recovery_efficiency: float = None,
        ach: float = None,
    ):
        super().__init__(name=name, category=SystemCategories.VENTILATION)
        self.zones = zones
        self.ach = ach
        self.fan_energy_coefficient = fan_energy_coefficient
        self.heat_recovery_efficiency = heat_recovery_efficiency

    def pre_process(self, idf: IDF):
        add_output_variable(
            idf=idf,
            key_values=self.zones,
            variables="Zone Mechanical Ventilation Standard Density Volume Flow Rate",
        )

        # Modify ACH if necessary
        if self.ach is not None:
            obj_name_arg = select_in_list(
                target_list=get_objects_name_list(
                    idf, "DesignSpecification:OutdoorAir"
                ),
                target=self.zones,
            )

            mod_fields = {
                "Outdoor_Air_Flow_Air_Changes_per_Hour": self.ach,
                "Outdoor_Air_Method": "AirChanges/Hour",
            }

            for field, value in mod_fields.items():
                energytool.base.idf_utils.set_named_objects_field_values(
                    idf=idf,
                    idf_object="DesignSpecification:OutdoorAir",
                    idf_object_names=obj_name_arg,
                    field_name=field,
                    values=value,
                )

        # Modify Heat Recovery if necessary
        if self.heat_recovery_efficiency is not None:
            obj_name_arg = select_in_list(
                target_list=get_objects_name_list(idf, "ZoneHVAC:IdealLoadsAirSystem"),
                target=self.zones,
            )

            mod_fields = {
                "Heat_Recovery_Type": "Sensible",
                "Sensible_Heat_Recovery_Effectiveness": self.heat_recovery_efficiency,
                "Latent_Heat_Recovery_Effectiveness": self.heat_recovery_efficiency,
            }
            for field, value in mod_fields.items():
                energytool.base.idf_utils.set_named_objects_field_values(
                    idf=idf,
                    idf_object="ZoneHVAC:IdealLoadsAirSystem",
                    idf_object_names=obj_name_arg,
                    field_name=field,
                    values=value,
                )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        air_volume = get_output_variable(
            eplus_res=eplus_results,
            key_values=self.zones,
            variables="Zone Mechanical Ventilation Standard Density Volume Flow Rate",
        )

        # Air_volume [m3/s] * 3600 [s] * fan_coef [Wh/m3] * 3600 [J/Wh]
        system_out = (air_volume * 3600 * self.fan_energy_coefficient * 3600).sum(
            axis=1
        )

        system_out.name = f"{self.name}_{Units.ENERGY.value}"
        return system_out.to_frame()


class DHWIdealExternal(System):
    """
    A model for simulating an ideal domestic hot water (DHW) system .
    This class represents an idealized DHW system. It allows you to model DHW energy
    consumption based on various parameters and on the number of occupants present in
    the zone(s).

    Parameters:
        name (str): The name of the DHW system.
        zones (str | List[str]): The name(s) of the zones where the DHW system is
            located.
        cop (float): The coefficient of performance (COP) for the DHW system, indicating
            its efficiency.
        t_dwh_set_point (float): The setpoint temperature for domestic hot water
            in degrees Celsius.
        t_cold_water (float): The temperature of the cold water supply in
            degrees Celsius.
        daily_volume_occupant (float): The daily volume of hot water consumed per
            occupant in liters.
        cp_water (float): The specific heat capacity of water in J/L·°C.

    Methods:
        pre_process(idf: IDF): pass.
        post_process(idf: IDF = None, eplus_results: pd.DataFrame = None)
            -> pd.DataFrame:
        Calculates DHW energy consumption and returns the results as a DataFrame.
    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        cop: float = 0.95,
        t_dwh_set_point: float = 60.0,
        t_cold_water: float = 15.0,
        daily_volume_occupant: float = 50.0,
        cp_water: float = 4183.2,
    ):
        super().__init__(name, category=SystemCategories.DHW)
        self.name = name
        self.zones = zones
        self.cop = cop
        self.t_dwh_set_point = t_dwh_set_point
        self.t_cold_water = t_cold_water
        self.daily_volume_occupant = daily_volume_occupant
        self.cp_water = cp_water

    def pre_process(self, idf: IDF):
        pass

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        nb_people = get_number_of_people(idf, zones=self.zones)

        # 4183.2[J/L.°C]
        daily_cons_per_occupant = (
            self.cp_water
            * (self.t_dwh_set_point - self.t_cold_water)
            * self.daily_volume_occupant
        )

        nb_days = eplus_results.resample("D").sum().shape[0]
        nb_entry = eplus_results.shape[0]

        dhw_consumption = daily_cons_per_occupant * nb_days * nb_people / self.cop

        return pd.DataFrame(
            {
                f"{self.name}_{Units.ENERGY.value}": (
                    np.ones(nb_entry) * dhw_consumption / nb_entry
                )
            },
            index=eplus_results.index,
        )


class ArtificialLighting(System):
    """
    A model for simulating artificial lighting systems energy consumption.

    Parameters:
        name (str): The name of the lighting system.
        zones (str | List[str]): The name(s) of the zones where the lighting system is
            present.
        power_ratio (float): The lighting power density in watts per square meter
            (W/m²).
        cop (float): The coefficient of performance (COP) for lighting system energy
            consumption (default is 1).

    Methods:
        pre_process(idf: IDF): Pre-processes the EnergyPlus IDF file to set
            lighting-related configurations.
        post_process(idf: IDF = None, eplus_results: pd.DataFrame = None)
            -> pd.DataFrame:
            Calculates lighting energy consumption and returns the results as a
            DataFrame.
    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        power_ratio: float = 3.0,
        cop: float = 1,
    ):  # W/m²
        super().__init__(name, category=SystemCategories.LIGHTING)
        self.name = name
        self.zones = zones
        self.power_ratio = power_ratio
        self.cop = cop

    def pre_process(self, idf: IDF):
        add_output_variable(
            idf=idf,
            key_values=self.zones,
            variables="Zone Lights Electricity Energy",
        )

        config = {
            "Design_Level_Calculation_Method": "Watts/Area",
            "Watts_per_Zone_Floor_Area": self.power_ratio,
        }
        obj_name_arg = select_in_list(
            target_list=get_objects_name_list(idf, "Lights"),
            target=self.zones,
        )

        for field, value in config.items():
            set_named_objects_field_values(
                idf=idf,
                idf_object="Lights",
                idf_object_names=obj_name_arg,
                field_name=field,
                values=value,
            )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        lighting_consumption = get_output_variable(
            eplus_res=eplus_results,
            key_values=self.zones,
            variables="Zone Lights Electricity Energy",
        )

        lighting_out = (lighting_consumption / self.cop).sum(axis=1)
        lighting_out.name = f"{self.name}_{Units.ENERGY.value}"
        return lighting_out.to_frame()


class AHUControl(System):
    """
    Represents an Air Handling Unit (AHU) control system for building energy modeling.
    This class is designed to model the control of an AHU system within a building
    energy model. It provides options for controlling the AHU based on either a
    predefined schedule or user-supplied data in the form of a Pandas DataFrame or
    Series.

    :param name: The name of the AHU control system.
    :param zones: The zones or spaces associated with the AHU control
        (default is "*," indicating all zones).
    :param control_strategy: The control strategy for the AHU, either "Schedule" or
        "DataFrame" (default is "Schedule").
    :param schedule_name: The name of the predefined schedule to use if the control
        strategy is "Schedule" (default is "ON_24h24h_FULL_YEAR").
    :param time_series: A Pandas DataFrame or Series containing user-defined control
        data (used when control_strategy is "DataFrame"). Default is None.

    :raises ValueError: If an invalid control strategy is specified.
    """

    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        control_strategy: str = "Schedule",
        schedule_name: str = "ON_24h24h_FULL_YEAR",
        time_series: pd.DataFrame | pd.Series = None,
    ):
        super().__init__(name, category=SystemCategories.VENTILATION)
        self.name = name
        self.zones = zones
        self.control_strategy = control_strategy
        self.schedule_name = schedule_name
        if time_series is not None:
            self.data_frame = as_1_column_dataframe(time_series)

    def pre_process(self, idf: IDF):
        if self.control_strategy == "Schedule":
            try:
                add_obj_from_obj_dict(
                    idf, RESOURCE_JSON, "Schedule:Compact".upper(), self.schedule_name
                )
            except ValueError:
                pass

        elif self.control_strategy == "DataFrame":
            add_hourly_schedules_from_df(idf, self.data_frame)
            self.schedule_name = self.data_frame.columns[0]

        else:
            raise ValueError("Specify valid control_strategy")

        # Get Design spec object to modify and set schedule
        obj_name_arg = select_in_list(
            target_list=get_objects_name_list(idf, "DesignSpecification:OutdoorAir"),
            target=self.zones,
        )

        set_named_objects_field_values(
            idf=idf,
            idf_object="DesignSpecification:OutdoorAir",
            idf_object_names=obj_name_arg,
            field_name="Outdoor_Air_Schedule_Name",
            values=self.schedule_name,
        )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        pass


class NaturalVentilation(System):
    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        ach=0.7,
        occupancy_schedule=True,
        ventilation_kwargs=None,
    ):
        super().__init__(name=name, category=SystemCategories.VENTILATION)
        self.name = name
        self.zones = zones
        self.ach = ach
        self.occupancy_schedule = occupancy_schedule
        self.ventilation_kwargs = ventilation_kwargs

    def pre_process(self, idf: IDF):
        add_natural_ventilation(
            idf,
            ach=self.ach,
            zones=self.zones,
            occupancy_schedule=self.occupancy_schedule,
            kwargs=self.ventilation_kwargs,
        )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        pass


class OtherEquipment(System):
    def __init__(
        self,
        name,
        zones: str | list = "*",
        distribute_load: bool = False,
        cop: float = 1.0,
        design_level_power: float = None,
        fraction_radiant: float = 0.2,
        compact_schedule_name: str = None,
        time_series: pd.Series = None,
        add_output_variables: bool = False,
    ):
        """
        This class is designed to model loads or heat source using other equipment
        systems within a building energy model. It provides options for specifying
        equipment parameters, distribution across zones, and scheduling methods.

        :param name: The name of the other equipment system.
        :param zones: The zones or spaces where the other equipment is located
            (default is "*," indicating all zones).
        :param distribute_load: If True, the equipment load is distributed based on
            zone areas. If False, the load is evenly distributed across specified zones
            (default is False).
        :param cop: The Coefficient of Performance (COP) for the equipment
            (default is 1.0). (eg. To model HP heating from compressor energy
            measurements)
        :param design_level_power: The design-level power of the equipment
            (if None, a predefined schedule is used for power levels).
        :param fraction_radiant: The fraction of radiant energy emitted by the
            equipment (default is 0.2).
        :param compact_schedule_name: The name of a predefined compact schedule to use
            for equipment operation (if None, a default schedule is used).
        :param time_series: A Pandas Series containing time-series data for equipment
            operation (if provided, it takes precedence over the compact_schedule_name).
        :param add_output_variables: If True, output variables for equipment heating
            energy are added to the EnergyPlus IDF (default is False).
        """
        super().__init__(name=name, category=SystemCategories.OTHER)
        self.cop = cop
        self.design_level_power = design_level_power
        self.add_output_variables = add_output_variables
        self.distribute_load = distribute_load
        self.fraction_radiant = fraction_radiant
        if time_series is not None:
            self.time_series = as_1_column_dataframe(time_series)
        else:
            self.time_series = None
        self.compact_schedule_name = compact_schedule_name
        self.schedule_name = None
        self.zones = zones

    def pre_process(self, idf: IDF):
        if self.zones == "*":
            self.zones = get_objects_name_list(idf, "Zone")
        else:
            self.zones = to_list(self.zones)

        # No time series was passed
        if self.time_series is None:
            # No compact schedule name is provided
            if self.compact_schedule_name is None:
                self.schedule_name = "ON_24h24h_FULL_YEAR"
                try:
                    add_obj_from_obj_dict(
                        idf,
                        RESOURCE_JSON,
                        "Schedule:Compact".upper(),
                        self.schedule_name,
                    )
                except ValueError:
                    pass

            # Compact schedule name is provided, but it can't be found
            elif not idf.getobject("Schedule:Compact", self.compact_schedule_name):
                raise ValueError(
                    f"{self.compact_schedule_name} not found inSchedule:Compact objects"
                )
            # Correct name has been given
            else:
                self.schedule_name = self.compact_schedule_name
        # Ime series was passed
        else:
            if self.compact_schedule_name:
                raise ValueError(
                    "Both compact_schedule_name and time_series were specified"
                )

            del_named_objects(idf, "Schedule:File", self.time_series.columns[0])
            add_hourly_schedules_from_df(idf=idf, data=self.time_series)
            self.schedule_name = self.time_series.columns[0]

        equipment_name_list = []
        if self.distribute_load:
            surf_arr = np.array([eppy.modeleditor.zonearea(idf, z) for z in self.zones])
            surf_ratio = surf_arr / np.sum(surf_arr)
        else:
            surf_ratio = np.array([1] * len(self.zones))

        for i, zone in enumerate(self.zones):
            equipment_name = f"{zone}_{self.name}_equipment"
            del_named_objects(idf, "OtherEquipment", equipment_name)
            equipment_name_list.append(equipment_name)

            idf.newidfobject(
                "OtherEquipment",
                Name=equipment_name,
                Zone_or_ZoneList_Name=zone,
                Schedule_Name=self.schedule_name,
                Design_Level_Calculation_Method="EquipmentLevel",
                Design_Level=surf_ratio[i] * self.design_level_power * self.cop,
                Fraction_Radiant=self.fraction_radiant,
            )

        if self.add_output_variables:
            add_output_variable(
                idf,
                key_values=equipment_name_list,
                variables="Other Equipment Total Heating Energy",
            )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        pass


class ZoneThermostat(System):
    def __init__(
        self,
        name: str,
        zones: str | list = "*",
        heating_compact_schedule_name: str = None,
        heating_time_series: pd.Series = None,
        cooling_compact_schedule_name: str = None,
        cooling_time_series: pd.Series = None,
        add_schedules_output_variables: bool = False,
        overwrite_heating_availability: bool = False,
        overwrite_cooling_availability: bool = False,
    ):
        """
        The ZoneThermostat class is designed to simplify the process of managing
        thermostat settings and schedule. It allows users to define thermostat
        configurations for specific zones or all zones in a building specifying
        compact schedule, or using Pandas Series.

        :param name: Name of the ZoneThermostat.
        :param zones: Zones to apply the thermostat settings to. "*" for all zones or
           a list of zone names.
        :param heating_compact_schedule_name: Name of the heating compact schedule.
           If not provided, and no time_series is given. The heating is artificially
            set to Off using a constant -60°C schedule.

        :param heating_time_series: Heating time series data as a pandas Series.
           If provided, this time series will be used for heating setpoint schedules,
            and any provided heating_compact_schedule_name will be ignored.

        :param cooling_compact_schedule_name: Name of the cooling compact schedule.
           If not provided, and no time_series is given. The cooling is artificially
            set to Off using a constant 100°C schedule.

        :param cooling_time_series: Cooling time series data as a pandas Series.
           If provided, this time series will be used for cooling setpoint schedules,
           and any provided cooling_compact_schedule_name will be ignored.

        :param add_schedules_output_variables: Whether to add schedules as output
           variables in the idf. This can be useful for analysis and
           visualization.

        :param overwrite_heating_availability: Whether to overwrite heating
           availability schedules for the specified zones.

        :param overwrite_cooling_availability: Whether to overwrite cooling availability
            schedules for the specified zones.
        """
        super().__init__(name=name, category=SystemCategories.OTHER)
        self.zones = zones
        self.add_schedules_output_variables = add_schedules_output_variables
        self.overwrite_heating_availability = overwrite_heating_availability
        self.overwrite_cooling_availability = overwrite_cooling_availability
        self.heating_compact_schedule_name = heating_compact_schedule_name
        self.heating_time_series = heating_time_series
        self.cooling_compact_schedule_name = cooling_compact_schedule_name
        self.cooling_time_series = cooling_time_series
        self.heating_schedule_name = None
        self.cooling_schedule_name = None

    def pre_process(self, idf: IDF):
        if self.zones == "*":
            self.zones = get_objects_name_list(idf, "Zone")
        else:
            self.zones = to_list(self.zones)

        ilas_list = get_zones_idealloadsairsystem(idf, self.zones)

        if self.heating_time_series is None:
            if self.heating_compact_schedule_name is None:
                self.heating_schedule_name = "-60C_heating_setpoint"
                try:
                    add_obj_from_obj_dict(
                        idf,
                        RESOURCE_JSON,
                        "Schedule:Compact".upper(),
                        self.heating_schedule_name,
                    )
                except ValueError:
                    pass

            elif not idf.getobject(
                "Schedule:Compact", self.heating_compact_schedule_name
            ):
                raise ValueError(
                    f"{self.heating_compact_schedule_name} not found in"
                    f"Schedule:Compact objects"
                )
            else:
                self.heating_schedule_name = self.heating_compact_schedule_name
        else:
            if self.heating_compact_schedule_name:
                raise ValueError("Both schedule name and time_series were specified")

            del_named_objects(idf, "Schedule:File", self.heating_time_series.name)

            add_hourly_schedules_from_df(
                idf=idf, data=as_1_column_dataframe(self.heating_time_series)
            )
            self.heating_schedule_name = self.heating_time_series.name

        if self.cooling_time_series is None:
            if self.cooling_compact_schedule_name is None:
                self.cooling_schedule_name = "100C_cooling_setpoint"
                try:
                    add_obj_from_obj_dict(
                        idf,
                        RESOURCE_JSON,
                        "Schedule:Compact".upper(),
                        self.cooling_schedule_name,
                    )
                except ValueError:
                    pass

            elif not idf.getobject(
                "Schedule:Compact", self.cooling_compact_schedule_name
            ):
                raise ValueError(
                    f"{self.cooling_compact_schedule_name} not found in"
                    f"Schedule:Compact objects"
                )
            else:
                self.cooling_schedule_name = self.cooling_compact_schedule_name
        else:
            if self.cooling_compact_schedule_name:
                raise ValueError(
                    "Both schedule name and time_series cannot be specified"
                )

            del_named_objects(idf, "Schedule:File", self.cooling_time_series.name)
            add_hourly_schedules_from_df(
                idf=idf, data=as_1_column_dataframe(self.cooling_time_series)
            )
            self.cooling_schedule_name = self.cooling_time_series.name

        if self.overwrite_heating_availability or self.overwrite_cooling_availability:
            self.cooling_schedule_name = "100C_cooling_setpoint"
            try:
                add_obj_from_obj_dict(
                    idf,
                    RESOURCE_JSON,
                    "Schedule:Compact".upper(),
                    "ON_24h24h_FULL_YEAR",
                )
            except ValueError:
                pass

        if self.overwrite_heating_availability:
            set_named_objects_field_values(
                idf=idf,
                idf_object="ZONEHVAC:IDEALLOADSAIRSYSTEM",
                field_name="Heating_Availability_Schedule_Name",
                idf_object_names=[ilas.Name for ilas in ilas_list],
                values="ON_24h24h_FULL_YEAR",
            )

        if self.overwrite_cooling_availability:
            set_named_objects_field_values(
                idf=idf,
                idf_object="ZONEHVAC:IDEALLOADSAIRSYSTEM",
                field_name="Cooling_Availability_Schedule_Name",
                idf_object_names=[ilas.Name for ilas in ilas_list],
                values="ON_24h24h_FULL_YEAR",
            )

        thermos_name_list = get_objects_name_list(
            idf, "ThermostatSetpoint:DualSetpoint"
        )

        thermos_to_keep = select_in_list(thermos_name_list, self.zones)

        set_named_objects_field_values(
            idf=idf,
            idf_object="ThermostatSetpoint:DualSetpoint",
            field_name="Heating_Setpoint_Temperature_Schedule_Name",
            idf_object_names=thermos_to_keep,
            values=self.heating_schedule_name,
        )

        set_named_objects_field_values(
            idf=idf,
            idf_object="ThermostatSetpoint:DualSetpoint",
            field_name="Cooling_Setpoint_Temperature_Schedule_Name",
            idf_object_names=thermos_to_keep,
            values=self.cooling_schedule_name,
        )

        if self.add_schedules_output_variables:
            add_output_variable(
                idf=idf,
                key_values=[self.heating_schedule_name, self.cooling_schedule_name],
                variables="Schedule Value",
            )

    def post_process(self, idf: IDF = None, eplus_results: pd.DataFrame = None):
        pass



import eppy
import pandas as pd
from eppy.modeleditor import IDF
from corrai.base.model import Model
from corrai.learning.sampling import expand_parameter_dict

from copy import deepcopy

from pathlib import Path

from eppy.runner.run_functions import run
import eppy.json_functions as json_functions
import enum

import energytool.base.idf_utils
from energytool.base.parse_results import read_eplus_res
from energytool.outputs import get_results
from energytool.system import System, SystemCategories
from energytool.base.idfobject_utils import (
    get_number_of_people,
    set_timestep,
    set_run_period,
)

import tempfile
import shutil
from contextlib import contextmanager

import platform
import os


class ParamCategories(enum.Enum):
    IDF = "idf"
    SYSTEM = "system"
    EPW_FILE = "epw_file"


class SimuOpt(enum.Enum):
    START = "start"
    STOP = "stop"
    TIMESTEP = "timestep"
    OUTPUTS = "outputs"
    EPW_FILE = "epw_file"
    VERBOSE = "verbose"


@contextmanager
def temporary_directory():
    if platform.system() == "Windows":
        user_home = os.path.expanduser("~")
        temp_path = os.path.join(user_home, r"AppData\Local\Temp")
    else:
        temp_path = None
    temp_dir = tempfile.mkdtemp(dir=temp_path)
    try:
        yield temp_dir

    finally:
        shutil.rmtree(temp_dir)


class Building(Model):
    """
    The Building class represents a building model. It is based on an EnergyPlus
    simulation file.

    :param idf_path: The path to the EnergyPlus IDF file that defines the building model.

    Attributes:
        idf: An EnergyPlus IDF object representing the building's configuration.
        systems: A dictionary that stores various categories of
        energytool HVAC systems associated with the building.

    Methods:
        set_idd(root_eplus): Sets the EnergyPlus IDD file used for parsing the IDF file.
        zone_name_list: Returns a list of names of zones defined in the building model.
        surface: Calculates and returns the total surface area of the building in square meters.
        volume: Calculates and returns the total volume of the building in cubic meters.
        add_system(system): Adds an HVAC system to the building's systems.
        del_system(system_name): Deletes an HVAC system from the building's systems by name.
        simulate(parameter_dict, simulation_options): Simulates the building model with
        specified parameters and simulation options, returning the simulation results
        as a pandas DataFrame.
    """

    def __init__(
        self,
        idf_path,
    ):
        self.idf = IDF(str(idf_path))
        self._idf_path = str(idf_path)
        self.systems = {category: [] for category in SystemCategories}

    @staticmethod
    def set_idd(root_eplus):
        try:
            IDF.setiddname(root_eplus / "Energy+.idd")
        except eppy.modeleditor.IDDAlreadySetError:
            pass

    @property
    def zone_name_list(self):
        return energytool.base.idf_utils.get_objects_name_list(self.idf, "Zone")

    @property
    def surface(self):
        return sum(
            eppy.modeleditor.zonearea(self.idf, z.Name)
            for z in self.idf.idfobjects["Zone"]
        )

    @property
    def volume(self):
        return sum(
            eppy.modeleditor.zonevolume(self.idf, z.Name)
            for z in self.idf.idfobjects["Zone"]
        )

    def __repr__(self):
        return f"""==Building==
Number of occupants: {round(get_number_of_people(self.idf), 2)}
Building surface: {self.surface} m²
Building volume: {self.volume} m3
Zone number: {len(self.zone_name_list)}

==HVAC systems==
Heating systems: {[obj.name for obj in self.systems[SystemCategories.HEATING]]}
Auxiliary: {[obj.name for obj in self.systems[SystemCategories.AUXILIARY]]}
Cooling systems: {[obj.name for obj in self.systems[SystemCategories.COOLING]]}
Ventilation system: {[obj.name for obj in self.systems[SystemCategories.VENTILATION]]}
Artificial lighting system: {[obj.name for obj in self.systems[SystemCategories.LIGHTING]]}
DHW production: {[obj.name for obj in self.systems[SystemCategories.DHW]]}
PV production: {[obj.name for obj in self.systems[SystemCategories.PV]]}
Sensors: {[obj.name for obj in self.systems[SystemCategories.SENSOR]]}
Others: {[obj.name for obj in self.systems[SystemCategories.OTHER]]}
"""

    def add_system(self, system: System):
        self.systems[system.category].append(system)

    def del_system(self, system_name: str):
        for cat in SystemCategories:
            for i, sys in enumerate(self.systems[cat]):
                if sys.name == system_name:
                    del self.systems[cat][i]

    def get_param_init_value(
        self,
        parameter_name_list: str | list[str] = None,
    ):
        """
        Returns the initial value(s) of one or more parameters of the model.

        :param parameter_name_list: A string or list of parameter names (str), like
            "idf.Material.SomeMat.Thickness" or "system.heating.Heater.cop"
        :return: a list of values if input is a list, or a single value if input is a string
        """
        if isinstance(parameter_name_list, str):
            is_single = True
            parameter_name_list = [parameter_name_list]
        else:
            is_single = False

        working_idf = deepcopy(self.idf)
        values = []

        for full_key in parameter_name_list:
            split_key = full_key.split(".")

            if split_key[0] == ParamCategories.IDF.value:
                value = energytool.base.idf_utils.getidfvalue(working_idf, full_key)
                values.append(value)

            elif split_key[0] == ParamCategories.SYSTEM.value:
                if split_key[1].upper() in [sys.value for sys in SystemCategories]:
                    sys_category = SystemCategories(split_key[1].upper())
                    system_obj = self.systems[sys_category][split_key[2]]
                    value = getattr(system_obj, split_key[3])
                    values.append(value)
                else:
                    raise ValueError(f"Unknown system category in key: {full_key}")

            elif split_key[0] == ParamCategories.EPW_FILE.value:
                values.append(str(self.weather.epw_path))

            else:
                raise ValueError(f"Unsupported parameter category in key: {full_key}")

        return values[0] if is_single else values

    def simulate(
        self,
        parameter_dict: dict[str, str | float | int] = None,
        simulation_options: dict[str, str | float | int] = None,
        idf_save_path: Path | None = None,
        param_mapping: dict = None,
    ) -> pd.DataFrame:
        """
        Simulate the building model with specified parameters and simulation options.

        :param parameter_dict: A dictionary containing key-value pairs representing
            parameters to be modified in the building model.
            These parameters can include changes to the IDF file, energytool HVAC system
            settings, or weather file.
            The key must represent the "path" to the parameter. "dots" must be separator.
            "idf" at the beginning of the path indicates a modification in the idf file
            "system" indicates a modification at energytool system level
            "epw_file" the path to epw file.
            see ParamCategories for allowed prefix

        :param simulation_options: A dictionary of simulation options that control
            the behavior of the EnergyPlus simulation.
            These options can include the choice of weather file, run period,
            time step, and desired outputs.
            See SimuOpt enum for allowed simulation options

        :param idf_save_path: (Optional) A Path where the modified
        IDF (Input Data File) will be saved after applying the specified
        parameter changes.
        If not provided, the modified IDF will not be saved separately.

        :return: A pandas DataFrame containing the simulation results, which may
            include energy consumption, indoor conditions, and other relevant data
            based on the specified outputs.

        The `simulate` method allows you to customize and run an EnergyPlus simulation
        for the building model. It provides the flexibility to modify various
        parameters and specify simulation options. It returns the results in a
        structured DataFrame

        Usage:
        # Example usage of the simulate method
        parameter_changes = {
            "idf.material.Urea Formaldehyde Foam_.1327.Conductivity": 0.05,
            "system.heating.Heater.cop": 0.5,
        }
        simulation_options = {
            'EPW_FILE': 'path/to/weather.epw',
            'START': '2023-01-01 00:00:00',
            'STOP': '2023-01-31 23:59:59',
            'TIMESTEP': 900
        }
        results = building.simulate(parameter_dict=parameter_changes,
        simulation_options=simulation_options)

        """
        working_idf = deepcopy(self.idf)
        working_syst = deepcopy(self.systems)

        epw_path = None
        if parameter_dict is None:
            parameter_dict = {}
        if param_mapping:
            parameter_dict = expand_parameter_dict(parameter_dict, param_mapping)

        for key in parameter_dict:
            split_key = key.split(".")

            # IDF modification
            if split_key[0] == ParamCategories.IDF.value:
                json_functions.updateidf(working_idf, {key: parameter_dict[key]})

            # In case it's a SYSTEM parameter, retrieve it in dict by category and name
            elif split_key[0] == ParamCategories.SYSTEM.value:
                if split_key[1].upper() in [sys.value for sys in SystemCategories]:
                    sys_key = SystemCategories(split_key[1].upper())
                else:
                    raise ValueError(
                        f"{split_key[1].upper()} is not part of SystemCategories"
                        f"choose one of {[elmt.value for elmt in SystemCategories]}"
                    )
                for syst in working_syst[sys_key]:
                    if syst.name == split_key[2]:
                        setattr(syst, split_key[3], parameter_dict[key])

            # Meteo file
            elif split_key[0] == ParamCategories.EPW_FILE.value:
                epw_path = parameter_dict[key]
            else:
                raise ValueError(
                    f"{split_key[0]} was not recognize as a valid parameter category"
                )

        # Simulation options
        if epw_path is None:
            try:
                epw_path = simulation_options[SimuOpt.EPW_FILE.value]
            except KeyError:
                raise ValueError(
                    "'epw_path' not found in parameter_dict nor in simulation_options"
                )
        elif SimuOpt.EPW_FILE.value in list(simulation_options.keys()):
            raise ValueError(
                "'epw_path' have been used in both parameter_dict and "
                "simulation_options"
            )
        ref_year = None
        if SimuOpt.START.value in simulation_options.keys():
            start = pd.to_datetime(simulation_options[SimuOpt.START.value])
            ref_year = start.year
            try:
                end = pd.to_datetime(simulation_options[SimuOpt.STOP.value])
            except KeyError:
                raise ValueError(
                    "Cannot set run period. Only start value was found "
                    "in simulation_options dict"
                )
            set_run_period(working_idf, start, end)

        if SimuOpt.TIMESTEP.value in simulation_options.keys():
            # timestep is supposed to be set in seconds
            set_timestep(
                working_idf,
                nb_timestep_per_hour=int(
                    3600 / simulation_options[SimuOpt.TIMESTEP.value]
                ),
            )

        # PRE-PROCESS
        system_list = [sys for sublist in working_syst.values() for sys in sublist]
        for system in system_list:
            system.pre_process(working_idf)

        # DEFAULT VERBOSE
        if SimuOpt.VERBOSE.value not in simulation_options.keys():
            simulation_options[SimuOpt.VERBOSE.value] = "v"

        # SIMULATE
        with temporary_directory() as temp_dir:
            working_idf.saveas((Path(temp_dir) / "in.idf").as_posix(), encoding="utf-8")
            idd_ref = working_idf.idd_version
            run(
                idf=working_idf,
                weather=epw_path,
                output_directory=temp_dir.replace("\\", "/"),
                annual=False,
                design_day=False,
                idd=None,
                epmacro=False,
                expandobjects=False,
                readvars=True,
                output_prefix=None,
                output_suffix=None,
                version=False,
                verbose=simulation_options[SimuOpt.VERBOSE.value],
                ep_version=f"{idd_ref[0]}-{idd_ref[1]}-{idd_ref[2]}",
            )

            eplus_res = read_eplus_res(
                Path(temp_dir) / "eplusout.csv", ref_year=ref_year
            )

            # Save IDF file after pre-process
            if idf_save_path:
                self.save(idf_save_path)

            # POST-PROCESS
            return get_results(
                idf=working_idf,
                eplus_res=eplus_res,
                systems=working_syst,
                outputs=simulation_options[SimuOpt.OUTPUTS.value],
            )

    def save(self, file_path: Path):
        """
        Save the current parameters of the model to a file.

        :param file_path: The file path where the parameters will be saved.
        """
        self.idf.saveas(file_path.as_posix(), encoding="utf-8")



import eppy
from eppy.modeleditor import IDF
import eppy.json_functions as json_functions

from energytool.tools import is_items_in_list, to_list


def getidfvalue(idf, param_key: str):
    """
    Get value from IDF object using a dotted key string, compatible with Eppy.
    Supports wildcard '*' for object name.
    """
    idftag, obj_type, obj_name, field = json_functions.key2elements(param_key)

    if idftag != "idf":
        raise ValueError(f"Unsupported prefix '{idftag}' in key: {param_key}")

    if obj_name == "*":
        idfobjs = idf.idfobjects[obj_type.upper()]
        return [obj[field] for obj in idfobjs if field in obj.fieldnames]
    else:
        obj = idf.getobject(obj_type.upper(), obj_name)
        if obj is None:
            raise KeyError(
                f"Object '{obj_name}' of type '{obj_type}' not found in IDF."
            )
        if field not in obj.fieldnames:
            raise KeyError(f"Field '{field}' not found in object '{obj_name}'.")
        return obj[field]


def get_objects_name_list(idf: IDF, idf_object: str):
    """
    Get a list of names of objects of a specified type in an EnergyPlus IDF file.

    :param idf: An EnergyPlus IDF object.
    :param idf_object: The name of the EnergyPlus object type for which to retrieve
        names.
    :return: A list of names of objects of the specified type in the IDF file.

    The function accesses objects of the given `idf_object` type in the IDF file and
    returns a list of their names.
    """
    return [obj.Name for obj in idf.idfobjects[idf_object]]


def get_named_objects(idf: IDF, idf_object: str, names: str | list):
    """
    Get specific named objects of a given type from an EnergyPlus IDF file.

    This function retrieves specific named objects of the specified EnergyPlus object
    type from an IDF file.

    :param idf: An EnergyPlus IDF object.
    :param idf_object: The name of the EnergyPlus object type from which to retrieve
        objects.
    :param names: The names of the specific objects to retrieve, either as a string
        or a list of strings.
    :return: A list of EnergyPlus objects that match the specified names.

    The function searches for objects of the given `idf_object` type in the IDF file and
    returns a list of objects whose names match the names provided in the `names`
    parameter.
    """
    names_list = to_list(names)
    objects_list = idf.idfobjects[idf_object]
    return [obj for obj in objects_list if obj.Name in names_list]


def get_building_surface_area(idf: IDF, outside_boundary_condition: str):
    """
    Return specific outside_boundary_condition building surface area
    """

    return sum(
        [
            o.area
            for o in idf.idfobjects["BuildingSurface:Detailed"]
            if o.Outside_Boundary_Condition == outside_boundary_condition
        ]
    )


def get_building_volume(idf: IDF):
    """Return volume based on zones volumes"""
    return sum(
        [
            eppy.modeleditor.zonevolume(idf, zname)
            for zname in get_objects_name_list(idf, "Zone")
        ]
    )


def is_value_in_objects_fieldname(
    idf: IDF,
    idf_object: str,
    field_name: str,
    values: float | int | str | list[float | int | str],
):
    """
    This function checks whether the specified values are present in the specified field
    of each instance of the specified EnergyPlus object type in the IDF file.

    :param idf: An EnergyPlus IDF object.
    :param idf_object: The name of the EnergyPlus object type to check.
    :param field_name: The name of the field within the object to check for values.
    :param values: The values to check for in the field.
    :return: A list of Boolean values, where each element corresponds to an instance
        of the specified EnergyPlus object type. True indicates that the value was found
        in the field, and False indicates that it was not found.

    For each instance of the `idf_object` in the IDF, this function checks if the
    specified `field_name` has the specified `values` as its variable value.

    Example:
    ```
    idf = IDF("example.idf")
    result = is_value_in_objects_fieldname(idf, "Zone", "ZoneName", ["Zone1", "Zone2"])
    # Returns [True, False] if "Zone1" is found in the "ZoneName" field of the first
    # instance and "Zone2" is not found in the field of the second instance.
    ```
    """
    values_list = to_list(values)
    obj_list = idf.idfobjects[idf_object]
    var_in_idf = [obj[field_name] for obj in obj_list]

    return is_items_in_list(var_in_idf, values_list)


def set_named_objects_field_values(
    idf: IDF,
    idf_object: str,
    field_name: str,
    values: str | float | list[str | float],
    idf_object_names: str | list = "*",
):
    """
    Set field values for one or more objects in an EnergyPlus IDF file.

    This function allows you to set the values of a specific field for multiple objects
    in an EnergyPlus IDF file. You can specify the objects and their names, the field
    name, and the values to set.

    :param idf: An EnergyPlus IDF object.
    :param idf_object: The name of the EnergyPlus object type for which to set values.
    :param field_name: The name of the field within the object to set.
    :param values: The value or list of values to set for the field.
    :param idf_object_names: (Optional) The names of the specific objects to update.
        By default, "*" is used to update all objects of the specified type.
    :return: None

    :raises ValueError: If the length of `values` and `idf_object_names` lists
        is not the same when updating multiple objects.
    """
    if idf_object_names == "*":
        idf_object_names_list = get_objects_name_list(idf, idf_object)
    else:
        idf_object_names_list = to_list(idf_object_names)

    values_list = to_list(values)
    if len(values_list) == 1:
        values_list = values_list * len(idf_object_names_list)

    if len(idf_object_names_list) != len(values_list):
        raise ValueError(
            "values and idf_object_names list must be of the "
            "same length. Or values must be a single object"
        )

    for obj_name, value in zip(idf_object_names_list, values_list):
        _set_named_object_field_value(idf, idf_object, obj_name, field_name, value)


def _set_named_object_field_value(
    idf, idf_object: IDF, name: str, field_name: str, value
):
    """
    Set the value of a field for a specific object in an EnergyPlus IDF file.

    This function allows you to set the value of a specific field for a particular
    object in an EnergyPlus IDF file. You need to specify the IDF object type, the
    object's name, the field name, and the value to set.

    :param idf: An EnergyPlus IDF object.
    :param idf_object: The name of the EnergyPlus object type to which the object belongs.
    :param name: The name of the specific object you want to update.
    :param field_name: The name of the field within the object to set.
    :param value: The value to set for the specified field.
    :return: None

    :raises ValueError: If the specified IDF object type is not found in the IDF file.
    """
    obj_list = idf.idfobjects[idf_object]
    if not obj_list:
        raise ValueError(f"{idf_object} not found in idf file")

    for obj in obj_list:
        if obj["Name"] == name:
            obj[field_name] = value


def get_named_objects_field_values(
    idf: IDF, idf_object: str, field_name: str, names: str | list = "*"
):
    """
    This function allows you to retrieve the values of a specific field for one or more
    named objects of a particular EnergyPlus object type in an IDF file.

    :param idf: An EnergyPlus IDF object.
    :param idf_object: The name of the EnergyPlus object type for which to retrieve values.
    :param field_name: The name of the field within the object from which to retrieve values.
    :param names: (Optional) The names of the specific objects for which to retrieve values.
        By default, "*" is used to retrieve values for all objects of the specified type.
    :return: A list of field values corresponding to the specified named objects.

    :raises ValueError: If the specified IDF object type is not found in the IDF file.
    """
    if names == "*":
        idf_object_names_list = get_objects_name_list(idf, idf_object)
    else:
        idf_object_names_list = to_list(names)

    return [
        _get_named_object_field_value(idf, idf_object, obj_name, field_name)
        for obj_name in idf_object_names_list
    ]


def _get_named_object_field_value(
    idf: IDF, idf_object: str, name: str, field_name: str
):
    obj_list = idf.idfobjects[idf_object]
    if not obj_list:
        raise ValueError(f"{idf_object} not found in idf file")

    for obj in obj_list:
        if obj.Name == name:
            return obj[field_name]


def del_named_objects(idf: IDF, idf_object: str, names: str | list = "*"):
    """
    Delete specific named objects of a given type from an EnergyPlus IDF file.

    This function allows you to remove one or more named objects of a specified
    EnergyPlus object type from an IDF file.

    :param idf: An EnergyPlus IDF object.
    :param idf_object: The name of the EnergyPlus object type from which to delete
        objects.
    :param names: (Optional) The names of the specific objects to delete.
        By default, "*"
        is used to delete all objects of the specified type.
    :return: None

    If the `names` parameter is set to "*", all objects of the specified type are
    deleted. Otherwise, the function removes only the named objects specified in
    the `names` list.
    """
    if names == "*":
        idf.idfobjects[idf_object] = []
    else:
        name_list = to_list(names)
        obj_to_remove = [idf.getobject(idf_object, name) for name in name_list]
        obj_list = idf.idfobjects[idf_object]

        idf.idfobjects[idf_object] = [o for o in obj_list if o not in obj_to_remove]


def copy_named_object_from_idf(
    source_idf: IDF, destination_idf: IDF, idf_object: str, name: str
):
    """
    This function allows you to copy a specific named object of a specified EnergyPlus
    object type from a source IDF file to a destination IDF file.

    :param source_idf: The source EnergyPlus IDF object from which to copy the object.
    :param destination_idf: The destination EnergyPlus IDF object to which the object
        should be copied.
    :param idf_object: The name of the EnergyPlus object type to which the object belongs.
    :param name: The name of the specific object to copy.
    :return: None

    The function retrieves the named object from the source IDF and copies it to the
    destination IDF if it is not already present in the destination IDF.

    Note: The function assumes that the object name is unique within the specified
    EnergyPlus object type.
    """
    # Get schedule in resources file
    obj_to_copy = source_idf.getobject(idf_object, name)

    # Copy in building idf if not already present
    destination_obj_list = destination_idf.idfobjects[idf_object]

    if obj_to_copy.Name not in get_objects_name_list(destination_idf, idf_object):
        destination_obj_list.append(obj_to_copy)



import datetime
import os
import tempfile
import uuid

import eppy
import numpy as np
import pandas as pd
from eppy.modeleditor import IDF

from energytool.base.idf_utils import (
    get_objects_name_list,
    is_value_in_objects_fieldname,
    get_building_surface_area,
    get_building_volume,
)
from energytool.tools import to_list, is_items_in_list


def idf_to_dict(idf: IDF):
    """
    Convert an IDF object into a nested dictionary representation.

    This function takes an IDF object (from eppy library)
    and converts it into a dictionary where each key is an IDF object
    type (e.g., 'BUILDING', 'MATERIAL'), and the value is another
    dictionary mapping the object name to a dictionary of its fields
    and values.

    Parameters
    ----------
    idf : IDF
        An instance of an IDF object containing IDFObjects with
        `fieldnames` and `fieldvalues` attributes.

    Returns
    -------
    dict
        A nested dictionary where the top-level keys are IDF object
        types, the second-level keys are object names, and the values
        are dictionaries mapping field names to their corresponding
        values.

    Examples
    --------
    >>> idf_dict = idf_to_dict(my_idf)
    >>> idf_dict["MATERIAL"]["Insulation"]["Thickness"]
    '0.1'
    """
    objects_dict = {}
    for obj in idf.idfobjects:
        if idf.idfobjects[obj]:
            objects_dict[obj] = {}
            for instance in idf.idfobjects[obj]:
                objects_dict[obj].update(
                    {
                        instance.Name: {
                            name: val
                            for name, val in zip(
                                instance.fieldnames, instance.fieldvalues
                            )
                        }
                    }
                )
    return objects_dict


def add_obj_from_obj_dict(idf: IDF, obj_dict: dict, idfobject: str, name: str):
    """
    Add an IDF object to an IDF instance from a nested object dictionary.

    This function adds a new IDF object to the given IDF instance,
    using data from a nested dictionary typically created by `idf_to_dict`.
    If an object with the same name already exists, a `ValueError` is raised.

    Parameters
    ----------
    idf : IDF
        An instance of an IDF object to which the new object will be added.
    obj_dict : dict
        A nested dictionary containing IDF object data. The format should
        match the output of `idf_to_dict` function, where the top-level keys are
        IDF object types, second-level keys are object names, and values
        are field dictionaries.
    idfobject : str
        The type of the IDF object to be added (e.g., 'MATERIAL', 'ZONE').
    name : str
        The name of the object to be added. Must be present in `obj_dict`.

    Raises
    ------
    ValueError
        If an object of the specified type and name already exists in the IDF.

    Examples
    --------
    >>> add_obj_from_obj_dict(idf, obj_dict, "MATERIAL", "NewInsulation")
    """
    if name not in get_objects_name_list(idf, idfobject):
        obj = obj_dict[idfobject][name]
        idf.newidfobject(**obj)
    else:
        raise ValueError(f"{idfobject}, named {name} already exists")


def get_zones_idealloadsairsystem(idf: IDF, zones: str | list = "*"):
    """
    Get a list of IdealLoadsAirSystem objects for specified zones in an EnergyPlus
    IDF file.

    :param idf: An EnergyPlus IDF object.
    :param zones: The zones for which to retrieve IdealLoadsAirSystem objects.
    This can be a single zone (string) or a list of zone names (list of strings).
    By default, "*" is used to retrieve IdealLoadsAirSystem objects for all zones.
    :return: A list of IdealLoadsAirSystem objects associated with the specified zones.

    The function first checks if the zones have HVAC equipment connections and then
    searches for IdealLoadsAirSystem objects associated with those zones.
    """
    if zones == "*":
        zones = get_objects_name_list(idf, "ZONE")
    else:
        zones = to_list(zones)

    ilas_list = []
    for zone in zones:
        equip_con = idf.getobject("ZONEHVAC:EQUIPMENTCONNECTIONS", zone)
        # If zone has hvac equipments
        if not equip_con:
            raise ValueError(f"{zone} doesn't have an IdealLoadAirSystem")

        equip_list = equip_con.get_referenced_object(
            "Zone_Conditioning_Equipment_List_Name"
        )
        for i in range(18):
            # 18 seem to be the max allowed (eppy)
            hvac_obj = equip_list.get_referenced_object(f"Zone_Equipment_{i + 1}_Name")
            if hvac_obj:
                if hvac_obj.key == "ZoneHVAC:IdealLoadsAirSystem":
                    ilas_list.append(hvac_obj)
    return ilas_list


def set_run_period(
    idf: IDF,
    simulation_start: datetime.datetime | pd.Timestamp,
    simulation_stop: datetime.datetime | pd.Timestamp,
):
    """
    Configure the IDF run period using datetime objects.

    This function allows you to set the IDF run period based on specified start and
    stop dates.

    :param idf: An EnergyPlus IDF object.
    :param simulation_start: The start date and time of the simulation as a
    datetime.datetime or pd.Timestamp object.
    :param simulation_stop: The stop date and time of the simulation as a
    datetime.datetime or pd.Timestamp object.
    :return: None

    The function configures the IDF run period with the provided start and stop dates,
    as well as other default settings for EnergyPlus simulation. Any existing run period
    configurations are cleared before adding the new run period definition.
    """

    run_period_list = idf.idfobjects["RunPeriod"]
    run_period_list.clear()
    idf.newidfobject(
        "RunPeriod",
        Name="run_period",
        Begin_Month=simulation_start.month,
        Begin_Day_of_Month=simulation_start.day,
        Begin_Year=simulation_start.year,
        End_Month=simulation_stop.month,
        End_Day_of_Month=simulation_stop.day,
        End_Year=simulation_stop.year,
        Day_of_Week_for_Start_Day=simulation_start.strftime("%A"),
        Use_Weather_File_Holidays_and_Special_Days="No",
        Use_Weather_File_Daylight_Saving_Period="No",
        Apply_Weekend_Holiday_Rule="Yes",
        Use_Weather_File_Rain_Indicators="Yes",
        Use_Weather_File_Snow_Indicators="Yes",
        Treat_Weather_as_Actual="No",
    )


def set_timestep(idf, nb_timestep_per_hour: int):
    timestep_list = idf.idfobjects["Timestep"]
    timestep_list.clear()
    idf.newidfobject("Timestep", Number_of_Timesteps_per_Hour=nb_timestep_per_hour)


def is_output_zone_variable(idf: IDF, zones: str, variables):
    """
    For the idfobject OUTPUT:VARIABLE, returns True if the required zone variable is
    already

    :param idf:
    :param zones:
    :param variables:
    :return:
    """
    zones_bool = is_value_in_objects_fieldname(
        idf, "Output:Variable", "Key_Value", zones
    )

    all_zones_bool = is_value_in_objects_fieldname(
        idf, "Output:Variable", "Key_Value", "*"
    )

    zones_bool = np.logical_or(zones_bool, all_zones_bool)

    variables_bool = is_value_in_objects_fieldname(
        idf, "Output:Variable", "Variable_Name", variables
    )

    return np.logical_and(zones_bool, variables_bool)


def del_output_zone_variable(idf, zones, variables):
    output_list = idf.idfobjects["OUTPUT:VARIABLE"]
    to_delete = is_output_zone_variable(idf, zones, variables)

    if np.any(to_delete):
        indices_to_remove = [i for i, trig in enumerate(to_delete) if trig]

        for idx in sorted(indices_to_remove, reverse=True):
            del output_list[idx]


def del_output_variable(idf, variables):
    output_list = idf.idfobjects["OUTPUT:VARIABLE"]
    to_delete = is_value_in_objects_fieldname(
        idf, "Output:Variable", "Variable_Name", variables
    )

    if np.any(to_delete):
        indices_to_remove = [i for i, trig in enumerate(to_delete) if trig]

        for idx in sorted(indices_to_remove, reverse=True):
            del output_list[idx]


def add_output_variable(
    idf: IDF,
    key_values: str | list,
    variables,
    reporting_frequency: str = "Hourly",
):
    """
    This function allows you to add output:variable object to an EnergyPlus IDF file.
    You can specify key values, variables, and reporting frequency for the output
    variables.

    :param idf: An EnergyPlus IDF object.
    :param key_values: The key values for which to add output variables.
        This can be a single key value (string) or a list of key values
        (list of strings).
    :param variables: The names of the variables to output. This can be a single
        variable name (string) or a list of variable names (list of strings).
    :param reporting_frequency: The reporting frequency for the output
        variables (e.g., "Hourly", "Daily", etc.). Default is "Hourly."
    :return: None

    The function iterates through the specified key values and variables, checking if
    corresponding output variables already exist in the IDF. If not, it adds new output
    variable definitions with the provided key values, variable names, and reporting
    frequency.

    Note: If a key value is set to "*", all existing output variables with the same
    variable name will be removed before adding the new definition.

    Example:
    ```
    idf = IDF("example.idf")
    add_output_variable(idf, "Zone1", "Zone Air Temperature")
    # Adds an output variable definition for "Zone Air Temperature" for "Zone1"
    # with default reporting frequency ("Hourly").
    ```
    """
    key_values_list = to_list(key_values)
    variables_list = to_list(variables)

    for key in key_values_list:
        for var in variables_list:
            if not np.any(is_output_zone_variable(idf, key, var)):
                if key == "*":
                    del_output_variable(idf, var)

                idf.newidfobject(
                    "OUTPUT:VARIABLE",
                    Key_Value=key,
                    Variable_Name=var,
                    Reporting_Frequency=reporting_frequency,
                )


def get_number_of_people(idf, zones="*"):
    zone_name_list = to_list(zones)
    if zones == "*":
        zone_list = idf.idfobjects["Zone"]
    else:
        zone_list = [idf.getobject("Zone", zname) for zname in zone_name_list]

    people_list = idf.idfobjects["People"]
    occupation = 0
    for zone in zone_list:
        try:
            # Compatibility with EnergyPlus version older than 2022
            if idf.idd_version[0] >= 22:
                people = next(
                    p
                    for p in people_list
                    if p.Zone_or_ZoneList_or_Space_or_SpaceList_Name == zone.Name
                )
            else:
                people = next(
                    p for p in people_list if p.Zone_or_ZoneList_Name == zone.Name
                )
        except StopIteration:
            continue

        people_method = people.Number_of_People_Calculation_Method
        if people_method == "People/Area":
            occupation += people.People_per_Zone_Floor_Area * zone.Floor_Area
        elif people_method == "People":
            occupation += people.Number_of_People
        elif people_method == "Area/Person":
            occupation += zone.Floor_Area / people.Zone_Floor_Area_per_Person
    return occupation


def add_hourly_schedules_from_df(
    idf: IDF,
    data: pd.DataFrame | pd.Series,
    schedule_type="Dimensionless",
    file_name=None,
    directory=None,
):
    """
    Add hourly schedules from a Pandas DataFrame or Series to an EnergyPlus IDF

    This function facilitates the integration of hourly schedule data into an
    EnergyPlus IDF, which is commonly used in building energy modeling.
    The provided data should represent hourly values for various parameters like
    temperature, occupancy, or lighting.

    :param idf: An EnergyPlus IDF object, which serves as the container for
        building simulation input data.
    :param data: A Pandas DataFrame or Series containing the hourly schedule data.
        The data should align with the EnergyPlus weather data format.
    :param schedule_type: The type of schedule data being added (e.g., "Dimensionless,"
        "Temperature," "Percent"). Default is "Dimensionless."
    :param file_name: The name of the CSV file where the schedule data will be
        temporarily stored before integration. If not provided, a random name will
         be generated.
    :param directory: The directory where the temporary CSV file will be stored.
        If not provided, a system-generated temporary directory will be used.

    Raises:
    - ValueError: If the input data is not a valid Pandas DataFrame or Series,
        or if it does not have the expected shape (8760 rows).
    - ValueError: If the schedule_type provided is not one of the valid EnergyPlus
        schedule types.
    - ValueError: If the length of schedule_type_list does not match the number
        of columns in the data.
    - ValueError: If schedule names in data columns already exist in the EnergyPlus IDF.

    Notes:
    The function reads the data, organizes it to match a single year
    (e.g., replacing years with 2009), and then writes it to a CSV file.
    Subsequently, it adds schedule objects to the EnergyPlus IDF,
    linking them to the CSV file.
    The schedules are defined as hourly data spanning 8760 hours,
    which corresponds to a typical year.
    """

    if isinstance(data, pd.Series):
        data = data.to_frame()
    if not isinstance(data, pd.DataFrame):
        raise ValueError("data must be a Pandas Series or DataFrame")
    if not (data.shape[0] == 8760 or data.shape[0] == 8760 + 24):
        raise ValueError("Invalid DataFrame. Dimension 0 must be 8760 or 8760 + 24")

    eplus_ref = [
        "Dimensionless",
        "Temperature",
        "DeltaTemperature",
        "PrecipitationRate",
        "Angle",
        "ConvectionCoefficient",
        "ActivityLevel",
        "Velocity",
        "Capacity",
        "Power",
        "Availability",
        "Percent",
        "Control",
        "Mode",
    ]

    schedule_type_list = to_list(schedule_type)
    if not np.array(
        is_items_in_list(items=schedule_type_list, target_list=eplus_ref)
    ).all():
        raise ValueError(
            f"f{schedule_type_list} is not a valid schedules type Valid types "
            f"are {eplus_ref}"
        )

    if len(schedule_type_list) == 1:
        schedule_type_list = schedule_type_list * len(data.columns)
    elif len(schedule_type_list) != len(data.columns):
        raise ValueError(
            "Invalid Schedule type list. Provide a single type"
            "or as many type as data columns"
        )

    already_existing = is_value_in_objects_fieldname(
        idf, idf_object="Schedule:File", field_name="Name", values=list(data.columns)
    )

    if np.array(already_existing).any():
        raise ValueError(
            f"{list(data.columns[already_existing])} already "
            f"presents in Schedules:Files"
        )

    if file_name is None:
        file_name = str(uuid.uuid4()) + ".csv"

    if directory is None:
        directory = tempfile.mkdtemp()

    full_path = os.path.realpath(os.path.join(directory, file_name))

    if len(data) != 8760 and len(data) != 8784:
        print(
            "Warning: the length of your data must either be 8760 or 8784. 8760 by Default"
        )
        number_hour = 8760
    else:
        number_hour = len(data)

    # In case we have data spanning over several years. Reorganise
    data.index = [idx.replace(year=2009) for idx in data.index]
    data.sort_index(inplace=True)
    data.to_csv(full_path, index=False, sep=",")

    for idx, (schedule, schedule_type) in enumerate(
        zip(data.columns, schedule_type_list)
    ):
        idf.newidfobject(
            "Schedule:File",
            Name=schedule,
            Schedule_Type_Limits_Name=schedule_type,
            File_Name=full_path,
            Column_Number=idx + 1,
            Rows_to_Skip_at_Top=1,
            Number_of_Hours_of_Data=number_hour,
            Column_Separator="Comma",
            Interpolate_to_Timestep="No",
        )


def add_natural_ventilation(
    idf: IDF,
    ach: float,
    zones: str | list = "*",
    occupancy_schedule: bool = True,
    minimum_indoor_temperature: float = 22.0,
    delta_temperature: float = 0,
    kwargs: dict = None,
):
    """
    This function facilitates the addition of natural ventilation settings to specific
    zones in an EnergyPlus IDF (Input Data File).
    Natural ventilation is modeled by specifying the desired Air Changes per Hour (ACH)
    for each zone. Users can also choose whether to link the ventilation schedule
    to occupancy or use a fixed schedule.

    :param idf: An EnergyPlus IDF object.
    :param ach: The desired Air Changes per Hour (ACH) for natural ventilation.
    :param zones: The zones to which natural ventilation settings should be applied.
        Can be a single zone name or a list of zone names. Default is "*," which
        applies the settings to all zones.
    :param occupancy_schedule: If True, the ventilation schedule is linked to occupancy
        schedules in the IDF. If False, a fixed schedule "On 24/7" is used for all
        specified zones (default is True).
    :param minimum_indoor_temperature: The minimum indoor temperature
        (in degrees Celsius) at which natural ventilation is allowed
        (default is 22.0°C).
    :param delta_temperature: The temperature difference (in degrees Celsius) above
        the outdoor temperature at which natural ventilation is initiated
        (default is 0.0°C).
    :param kwargs: Additional properties for "ZoneVentilation:DesignFlowrate" object
    """

    if kwargs is None:
        kwargs = {}

    if zones == "*":
        z_list = get_objects_name_list(idf, "Zone")
    else:
        z_list = to_list(zones)

    if occupancy_schedule:
        zone_sched_dict = {}
        for ppl in idf.idfobjects["People"]:
            z_name = ppl.get_referenced_object("Zone_or_ZoneList_Name").Name
            if z_name in z_list:
                zone_sched_dict[z_name] = ppl.Number_of_People_Schedule_Name
    else:
        if not idf.getobject("Schedule:Compact", "On 24/7"):
            idf.newidfobject(
                key="Schedule:Compact",
                Name="On 24/7",
                Schedule_Type_Limits_Name="Any Number",
                Field_1="Through: 12/31",
                Field_2="For: AllDays",
                Field_3="Until: 24:00",
                Field_4=1,
            )
        zone_sched_dict = {z_name: "On 24/7" for z_name in z_list}

    for z_name in zone_sched_dict.keys():
        vnat = idf.getobject("ZoneVentilation:DesignFlowrate", f"Natvent_{z_name}")

        if vnat:
            idf.idfobjects["ZoneVentilation:DesignFlowrate"].remove(vnat)

        idf.newidfobject(
            "ZoneVentilation:DesignFlowrate",
            Name=f"Natvent_{z_name}",
            Zone_or_ZoneList_Name=z_name,
            Schedule_Name=zone_sched_dict[z_name],
            Design_Flow_Rate_Calculation_Method="AirChanges/Hour",
            Design_Flow_Rate=ach,
            Minimum_Indoor_Temperature=minimum_indoor_temperature,
            Delta_Temperature=delta_temperature,
            **kwargs,
        )


def get_n50_from_q4(q4, heated_volume, outside_surface, n=2 / 3):
    """
    Outside surface correspond to building surface in contact with outside Air
    n is flow exponent. 1 is laminar 0.5 is turbulent. Default 2/3
    """
    return q4 / ((4 / 50) ** n * heated_volume / outside_surface)


def get_ach_from_n50(n50, delta_qv, wind_exposition=0.07, f=15):
    """
    wind_exposition : ranging from 0.1 to 0.04 default 0.07
    f : can't remember why but default is 15
    delta_qv = in ach , the difference between mechanically blown and extracted
        air.
    For extraction only  delta_qv = Qv, for crossflow ventilation delta_qv = 0

    """
    return n50 * wind_exposition / (1 + f / wind_exposition * (delta_qv / n50) ** 2)


def get_building_infiltration_ach_from_q4(idf, q4pa=1.2, wind_exposition=0.07, f=15):
    building_outdoor_surface = get_building_surface_area(
        idf, outside_boundary_condition="Outdoors"
    )
    building_volume = get_building_volume(idf)

    # Compute N50 from q4pa
    n50 = get_n50_from_q4(
        q4=q4pa, heated_volume=building_volume, outside_surface=building_outdoor_surface
    )

    # Get qv
    z_ach_dict = {}
    for siz in idf.idfobjects["Sizing:Zone"]:
        zone = siz.get_referenced_object("Zone_or_ZoneList_Name")
        design = siz.get_referenced_object(
            "Design_Specification_Outdoor_Air_Object_Name"
        )
        if design.Outdoor_Air_Method != "AirChanges/Hour":
            raise ValueError(
                "Outdoor Air method other than AirChanges/Hour not yet implemented"
            )
        z_ach_dict[zone.Name] = design.Outdoor_Air_Flow_Air_Changes_per_Hour

    z_hx_dict = {}
    for connection in idf.idfobjects["ZoneHVAC:EquipmentConnections"]:
        z_name = connection.Zone_Name
        sys = connection.get_referenced_object(
            "Zone_Conditioning_Equipment_List_Name"
        ).get_referenced_object("Zone_Equipment_1_Name")
        z_hx_dict[z_name] = sys.Heat_Recovery_Type

    qv = (
        sum(
            [
                eppy.modeleditor.zonevolume(idf, zname) * z_ach_dict[zname]
                for zname in z_ach_dict.keys()
                if z_hx_dict[zname] == "None"
            ]
        )
        / building_volume
    )

    return get_ach_from_n50(n50, delta_qv=qv, wind_exposition=wind_exposition, f=f)


def get_windows_by_boundary_condition(idf, boundary_condition):
    ext_surf_name = [
        obj.Name
        for obj in idf.idfobjects["BuildingSurface:Detailed"]
        if obj.Outside_Boundary_Condition == boundary_condition
    ]

    return [
        obj
        for obj in idf.idfobjects["FenestrationSurface:Detailed"]
        if obj.Building_Surface_Name in ext_surf_name
    ]


def get_constructions_layer_list(constructions):
    construction_list = to_list(constructions)
    material_name_list = []
    for constructions in construction_list:
        material_name_list += [
            elmt
            for elmt, key in zip(constructions.fieldvalues, constructions.fieldnames)
            if key not in ["key", "Name"]
        ]
    return material_name_list


def del_layer_from_constructions(building, names):
    names_list = to_list(names)

    new_cons_list = []
    for construction in building.idf.idfobjects["Construction"]:
        keys = [k for k in construction.fieldnames]
        values = [v for v in construction.fieldvalues if v not in names_list]

        new_cons = {k: v for v, k in zip(values, keys)}

        new_cons_list.append(new_cons)

    building.idf.idfobjects["Construction"] = [
        building.idf.newidfobject(**cons) for cons in new_cons_list
    ]


def idf_object_to_dict(obj):
    return {k: v for k, v in zip(obj.fieldnames, obj.fieldvalues)}



import enum

import pandas as pd
from eppy.modeleditor import IDF

from energytool.base.units import Units
from energytool.system import System, SystemCategories


class OutputCategories(enum.Enum):
    RAW = "RAW"
    SYSTEM = "SYSTEM"
    SENSOR = "SENSOR"


def get_results(
    idf: IDF,
    eplus_res: pd.DataFrame,
    outputs: str,
    systems: dict[SystemCategories, list[System]] = None,
):
    """
    Retrieve HVAC systems results based on specified output categories.

    :param eplus_res: DataFrame containing EnergyPlus simulation results.
    :param outputs: String containing pipe-separated output categories
        (e.g., "RAW|SYSTEM"). Categories must be values from OutputCategories enum
    :param idf: IDF object representing the EnergyPlus input data.
    :param systems: Optional, dictionary mapping SystemCategories to lists of System
        objects.
    :return: A DataFrame containing the concatenated results based on the specified
        categories.
    """
    to_return = []
    split_outputs = outputs.split("|")
    for output_cat in split_outputs:
        if output_cat == OutputCategories.RAW.value:
            to_return.append(eplus_res)
        elif output_cat == OutputCategories.SYSTEM.value:
            results = get_system_energy_results(idf, systems, eplus_res)
            if results is not None:
                to_return.append(results)
        elif output_cat == OutputCategories.SENSOR.value:
            results = get_sensor_results(idf, systems, eplus_res)
            if results is not None:
                to_return.append(results)

        else:
            raise ValueError(f"{output_cat} not recognized or not yet implemented")

    if to_return:
        concatenated = pd.concat(to_return, axis=1)
        concatenated = concatenated.loc[
            :, ~concatenated.columns.duplicated()
        ]  # to avoid duplicates

        return concatenated


def get_sensor_results(
    idf: IDF,
    systems: dict[SystemCategories, list[System]],
    eplus_res: pd.DataFrame,
):
    result_list = []
    for sens in systems[SystemCategories.SENSOR]:
        result_list.append(sens.post_process(idf, eplus_res))
    return pd.concat(result_list, axis=1)


def get_system_energy_results(
    idf: IDF,
    systems: dict[SystemCategories, list[System]],
    eplus_res: pd.DataFrame,
):
    """
    Retrieve energy results for systems contains in the SystemCategories.
    If several systems are present in a category, it will return the sum of there
    energy use.
    The energy absorbed by a system is identified by the ENERGY_[J] tag in its name.
    A TOTAL_ENERGY_[J] column sums all the energy consumed by the systems

    :param idf: IDF object representing the EnergyPlus input data.
    :param systems: Dictionary mapping SystemCategories to lists of System objects.
    :param eplus_res: DataFrame containing EnergyPlus simulation results.
    :return: A DataFrame containing energy results for different system categories.
    """
    sys_nrj_res = []
    hvac_list = [
        SystemCategories.HEATING,
        SystemCategories.COOLING,
        SystemCategories.AUXILIARY,
        SystemCategories.DHW,
        SystemCategories.VENTILATION,
        SystemCategories.LIGHTING,
    ]

    for cat in hvac_list:
        syst_list = systems[cat]
        if syst_list:
            cat_res = []
            for system in syst_list:
                res = system.post_process(idf, eplus_results=eplus_res)
                if res is not None:
                    unit = Units.ENERGY.value
                    unit = unit.replace("[", r"\[").replace("]", r"\]")
                    cat_res.append(
                        res.loc[:, res.columns.str.contains(unit, regex=True)]
                    )
            if cat_res:
                cat_res_series = pd.concat(cat_res, axis=1).sum(axis=1)
                cat_res_series.name = f"{cat.value}_{Units.ENERGY.value}"
                sys_nrj_res.append(cat_res_series)

    if sys_nrj_res:
        sys_nrj_res_df = pd.concat(sys_nrj_res, axis=1)
        sys_nrj_res_df[f"TOTAL_SYSTEM_{Units.ENERGY.value}"] = sys_nrj_res_df.sum(
            axis=1
        )
        return sys_nrj_res_df
